import base64
import concurrent
import copy
import datetime as dt
import functools
import ipaddress
import json
import logging
import math
import os
import pickle
import random
import re
import shlex
import time
import traceback
import typing as t
import uuid
from collections import OrderedDict

from flask import request, current_app, jsonify, g
from flask_jwt_extended import jwt_required, get_jwt_identity, create_access_token
from pkg_resources import parse_version

import dimensigon
import dimensigon.use_cases.clustering
import dimensigon.use_cases.routing as routing
import dimensigon.web.network as ntwrk
from dimensigon import defaults as d, defaults
from dimensigon.domain.entities import Software, Server, SoftwareServerAssociation, Catalog, Route, StepExecution, \
    Orchestration, OrchExecution, User, ActionTemplate, ActionType, Gate
from dimensigon.domain.entities.route import RouteContainer
from dimensigon.network.auth import HTTPBearerAuth
from dimensigon.utils import asyncio, subprocess
from dimensigon.utils.dag import DAG
from dimensigon.utils.event_handler import Event
from dimensigon.utils.helpers import get_distributed_entities, is_iterable_not_string, md5, get_now, format_exception
from dimensigon.utils.var_context import VarContext
from dimensigon.web import db, executor, errors, threading
from dimensigon.web.api_1_0 import api_bp
from dimensigon.web.async_functions import deploy_orchestration, async_send_file
from dimensigon.web.background_tasks import _async_get_neighbour_catalog_data_mark, \
    upgrade_version, \
    update_catalog
from dimensigon.web.decorators import securizer, forward_or_dispatch, validate_schema, lock_catalog, log_time
from dimensigon.web.helpers import check_param_in_uri, normalize_hosts, search, get_auth_from_request
from dimensigon.web.json_schemas import launch_command_post, routes_post, routes_patch, \
    launch_orchestration_post, send_post, orchestration_full, manager_server_ignore_lock_post, cluster_post

if t.TYPE_CHECKING:
    from dimensigon.use_cases.operations import IOperationEncapsulation


@api_bp.route('/')
def home():
    return "API v1.0 documentation page"


@api_bp.route('/join/token', methods=['GET'])
@jwt_required
def join_token():
    if get_jwt_identity() == '00000000-0000-0000-0000-000000000001':
        user_join = User.get_by_user('join')
        expire_time = request.args.get('expires_time', type=int, default=None)

        kwargs = {}
        if expire_time:
            kwargs['expires_delta'] = dt.timedelta(minutes=expire_time)

        return {'token': create_access_token(str(user_join.id), **kwargs)}, 200
    else:
        return {}, 401


@api_bp.route('/join/public', methods=['GET'])
@jwt_required
def join_public():
    if User.query.get(get_jwt_identity()) == User.get_by_user('join'):
        return g.dimension.public.save_pkcs1(), 200, {'content-type': 'application/octet-stream'}

    else:
        return {}, 401


servers_to_be_created = {}


def delete_old_temp_servers():
    global servers_to_be_created
    now = get_now()
    for s_id, data in list(servers_to_be_created.items()):
        s_created = dt.datetime.strptime(data['created_on'], defaults.DATETIME_FORMAT)
        if now - s_created > dt.timedelta(hours=1):
            servers_to_be_created.pop(s_id, None)


fetched_catalog = (get_now(), None)
_lock = threading.Lock()


@api_bp.route('/join', methods=['POST'])
@securizer
@jwt_required
def join():
    global fetched_catalog
    if get_jwt_identity() == '00000000-0000-0000-0000-000000000004':
        js = request.get_json()
        current_app.logger.debug(f"New server wanting to join: {json.dumps(js, indent=2)}")
        if db.session.query(Server).filter_by(id=js.get('id', None)).count() > 0:
            raise errors.DuplicatedId(js.get('id', None))
        if db.session.query(Server).filter_by(name=js.get('name', None)).count() > 0:
            raise errors.AlreadyExists('name', js.get('name', None))
        s = Server.from_json(js)
        s.created_on = get_now()
        external_ip = ipaddress.ip_address(request.remote_addr)
        if not external_ip.is_loopback and external_ip not in [gate.ip for gate in s.gates]:
            for port in set([gate.port for gate in s.gates]):
                s.add_new_gate(external_ip, port, hidden=True)

        certfile = current_app.dm.config.http_conf.get('certfile', None)
        keyfile = current_app.dm.config.http_conf.get('keyfile', None)

        if keyfile and os.path.exists(keyfile):
            with open(keyfile, 'rb') as fh:
                keyfile_content = fh.read()
        else:
            raise errors.FileNotFound(keyfile)
        if certfile and os.path.exists(certfile):
            with open(certfile, 'rb') as fh:
                certfile_content = fh.read()
        else:
            raise errors.FileNotFound(certfile)

        data = {'keyfile': base64.b64encode(keyfile_content).decode(),
                'certfile': base64.b64encode(certfile_content).decode()}

        data.update(Dimension=g.dimension.to_json())
        data.update(me=str(Server.get_current().id))

        with _lock:
            if fetched_catalog[1] is None or fetched_catalog[0] < get_now() - dt.timedelta(minutes=1):
                c = fetch_catalog(defaults.INITIAL_DATEMARK)
                fetched_catalog = (get_now(), c)
            else:
                c = fetched_catalog[1]
        data.update(catalog=c)

        server_data = s.to_json(add_gates=True)
        servers_to_be_created.update({s.id: server_data})
        del s
        return data, 200
    else:
        raise errors.GenericError('Invalid token', status_code=400)


_lock_add_node = threading.Lock()

@api_bp.route('/join/acknowledge/<server_id>', methods=['POST'])
@jwt_required
@lock_catalog
def join_acknowledge(server_id):
    server_data = servers_to_be_created.get(server_id, None)
    if not server_data:
        raise errors.EntityNotFound('Server', server_id)
    s = Server.from_json(server_data)
    with _lock_add_node:
        db.session.add(s)
        db.session.commit()
        current_app.logger.debug(f"Server join acknowledge {s.name}")
    return {}, 204


@api_bp.route('/manager/server_ignore_lock', methods=['POST'])
@forward_or_dispatch()
@jwt_required
@securizer
@validate_schema(manager_server_ignore_lock_post)
def internal_server():
    if get_jwt_identity() == '00000000-0000-0000-0000-000000000001':
        ignore = request.get_json()['ignore_on_lock']
        for server_id in request.get_json()['server_ids']:
            server = Server.query.get_or_raise(server_id)
            server.l_ignore_on_lock = ignore
            db.session.commit()

        return {}, 204
    else:
        raise errors.UserForbiddenError


@api_bp.route('/send', methods=['POST'])
@forward_or_dispatch()
@jwt_required
@securizer
@validate_schema(send_post)
def send():
    def search_cost(ssa, route_list):
        cost = [route['cost'] for route in route_list if str(ssa.server.id) == route['destination_id']]
        if cost:
            if cost[0] is None:
                cost = 999999
            else:
                cost = cost[0]
        else:
            cost = 999999
        return cost

    auth = HTTPBearerAuth(create_access_token(get_jwt_identity(), expires_delta=dt.timedelta(
        minutes=30)))
    # Validate Data
    json_data = request.get_json()

    dest_server = Server.query.get_or_raise(json_data['dest_server_id'])

    if 'software_id' in json_data:
        software = Software.query.get_or_raise(json_data['software_id'])

        ssa = SoftwareServerAssociation.query.filter_by(server=g.server, software=software).one_or_none()
        # if current server does not have the software, forward request to the closest server who has it
        if not ssa:
            resp = ntwrk.get(dest_server, 'api_1_0.routes', auth=auth, timeout=5)
            if resp.code == 200:
                ssas = copy.copy(software.ssas)
                ssas.sort(key=functools.partial(search_cost, route_list=resp.msg['route_list']))
            # unable to get route cost, we take the first option we have
            else:
                ssas = random.shuffle(list(software.ssas))
            if len(ssas) == 0:
                raise errors.NoSoftwareServer(software_id=str(software.id))
            server = ssas[0].server  # closest server from dest_server who has the software

            resp = ntwrk.post(server, 'api_1_0.send', json=json_data, auth=auth)
            resp.raise_if_not_ok()
            return resp.msg, resp.code
        else:

            file = os.path.join(ssa.path, software.filename)
            if not os.path.exists(file):
                raise errors.FileNotFound(file)
            size = ssa.software.size
    else:
        file = json_data['file']
        if os.path.exists(file):
            size = os.path.getsize(file)
            checksum = md5(json_data.get('file'))
        else:
            raise errors.FileNotFound(file)

    chunk_size = min(json_data.get('chunk_size', d.CHUNK_SIZE), d.CHUNK_SIZE) * 1024
    max_senders = min(json_data.get('max_senders', d.MAX_SENDERS), d.MAX_SENDERS)
    chunks = math.ceil(size / chunk_size)

    if 'software_id' in json_data:
        json_msg = dict(software_id=str(software.id), num_chunks=chunks,
                        dest_path=json_data.get('dest_path', current_app.config['SOFTWARE_REPO']))
    else:
        json_msg = dict(filename=os.path.basename(json_data.get('file')), size=size, checksum=checksum,
                        num_chunks=chunks, dest_path=json_data.get('dest_path', current_app.config['SOFTWARE_REPO']))
    if 'force' in json_data:
        json_msg['force'] = json_data['force']

    resp = ntwrk.post(dest_server, 'api_1_0.transferlist', json=json_msg, auth=auth)
    resp.raise_if_not_ok()

    transfer_id = resp.msg.get('id')
    current_app.logger.debug(
        f"Transfer {transfer_id} created. Sending {file} to {dest_server}:{json_data.get('dest_path')}.")

    if json_data.get('background', True):
        executor.submit(asyncio.run,
                        async_send_file(dest_server=dest_server, transfer_id=transfer_id, file=file,
                                        chunk_size=chunk_size, max_senders=max_senders, auth=auth))
    else:
        asyncio.run(async_send_file(dest_server=dest_server, transfer_id=transfer_id, file=file,
                                    chunk_size=chunk_size, max_senders=max_senders, auth=auth))

    if json_data.get('include_transfer_data', False):
        resp = ntwrk.get(dest_server, "api_1_0.transferresource", view_data=dict(transfer_id=transfer_id), auth=auth)
        if resp.code == 200:
            msg = resp.msg
        else:
            resp.raise_if_not_ok()
    else:
        msg = {'transfer_id': transfer_id}
    return msg, 202 if json_data.get('background', True) else 201


@api_bp.route('/software/dimensigon', methods=['GET'])
@forward_or_dispatch()
@jwt_required
@securizer
def software_dimensigon():
    # sends the last software
    repo = os.path.join(current_app.dm.config.config_dir, defaults.SOFTWARE_REPO, defaults.DIMENSIGON_DIR)
    max_version = None
    max_file = None
    for file in os.listdir():
        if 'dimensigon-' in file:
            m = re.search(r'v?\d+\.\d+[.-][ab]?\d+', file)
            if m and parse_version(m.group()) >= parse_version(max_version or dimensigon.__version__):
                max_version = m.group()
                max_file = file
    if max_file:
        with open(os.path.join(repo, 'dimensigon', max_file), 'rb') as fh:
            return {'filename': max_file, 'version': max_version,
                    'content': base64.b64encode(fh.read()).decode('ascii')}, 200
    else:
        return {}, 204


@api_bp.route('/catalog', methods=['POST'])
@forward_or_dispatch()
@jwt_required
@securizer
def catalog_update():

    data = asyncio.run(_async_get_neighbour_catalog_data_mark(get_now().strftime(defaults.DATETIME_FORMAT)))
    # check version upgrade before catalog upgrade to match database revision
    if not upgrade_version(data):
        update_catalog(data)
    db.session.commit()
    return {}, 204


@api_bp.route('/catalog/<string:data_mark>', methods=['GET', 'POST'])
@forward_or_dispatch()
@jwt_required
@securizer
def catalog(data_mark):
    # Input Validation
    if data_mark == 'initial':
        data_validated = dt.datetime(dt.MINYEAR, 1, 1)
    else:
        try:
            data_validated = dt.datetime.strptime(data_mark, defaults.DATEMARK_FORMAT)
        except Exception as e:
            return {'error': f'Invalid Data Mark: {e}'}, 400

    return fetch_catalog(data_validated)


@log_time()
def fetch_catalog(data_mark):
    data = {}
    now = get_now()
    for name, obj in get_distributed_entities():
        c = Catalog.query.get(name)
        # db.session.query to bypass deleted objects to spread deleted changes
        repo_data = obj.query.filter(obj.last_modified_at > data_mark).filter(
            obj.last_modified_at <= now).all()
        if name == 'User':
            data.update({name: [e.to_json(password=True) for e in repo_data]})
        else:
            data.update({name: [e.to_json() for e in repo_data]})
    return data


_cluster_logger = logging.getLogger('dimensigon.cluster')


@api_bp.route('/cluster', methods=['POST'])
@log_time('full')
# @forward_or_dispatch()
@jwt_required
@securizer
@log_time('after securizer')
@validate_schema(cluster_post)
@log_time('after validation')
def cluster():
    if get_jwt_identity() == '00000000-0000-0000-0000-000000000001':
        data = request.get_json()
        # _cluster_logger.debug(f"Data received {clustering.log_data(data)}")
        current_app.cluster_manager.put(data)
        return {}, 204
    else:
        raise errors.UserForbiddenError

async def background_cluster_in(server_id, routes, auth):
    server = Server.query.get(server_id)
    servers = Server.get_neighbours(exclude=server)
    # cluster information
    # if cr:
    #     tasks.append(clustering.send_cluster_register(cr, servers=servers, auth=auth))

    # route information
    changed_routes = {}

    # server might be sent cluster in message but not created in database
    if server:
        new_route = await routing.async_check_gates(server, timeout=5, retries=3, delay=2)
        if new_route and isinstance(new_route, RouteContainer):
            routing.logger.debug(f'cluster IN: New neighbour {server} found through {new_route.gate}')
            routing.set_route(server, new_route)
            changed_routes.update({server: new_route})
        else:
            routing.logger.debug(f"cluster IN: {server} is not a neighbour")
    try:
        changed_routes.update(
            routing.update_route_table_from_data({'server_id': server_id, 'route_list': routes}, auth))
    except errors.InvalidRoute as e:
        debug_new_routes = []
        routes.sort(key=lambda x: x.get('cost') or routing.MAX_COST, reverse=True)
        for new_route in routes:
            target_server = Server.query.get(new_route.get('destination_id'))
            proxy_server = Server.query.get(new_route.get('proxy_server_id'))
            gate = Gate.query.get(new_route.get('gate_id'))
            debug_new_route = dict(destination=getattr(target_server, 'name', new_route.get('destination_id')),
                                   proxy_server=getattr(proxy_server, 'name', new_route.get('proxy_server_id')),
                                   gate=str(gate) if gate else new_route.get('gate_id'),
                                   cost=new_route.get('cost'))
            debug_new_routes.append(debug_new_route)
            routing.logger.exception(
                "Error setting routes from following data: " + json.dumps(debug_new_routes, indent=4))
        raise

    if changed_routes:
        await routing.async_send_routes(changed_routes, auth=auth, servers=servers, exclude=server)


@api_bp.route('/cluster/in/<server_id>', methods=['POST'])
@jwt_required
@securizer
def cluster_in(server_id):
    user = User.get_current()
    if user and user.user == 'root':
        cr = current_app.cluster_manager.set_alive(server_id)
        _cluster_logger.debug(
            f"{getattr(Server.query.get(server_id), 'name', server_id) or server_id} is a new alive server")

        # run background execution to check routes
        executor.submit(asyncio.run, background_cluster_in(server_id, request.get_json(), get_auth_from_request()))

        return {'cluster': current_app.cluster_manager.cluster.get_cluster(),
                'neighbours': [s.id for s in Server.get_neighbours()]}, 200

    else:
        raise errors.UserForbiddenError


async def background_cluster_out(server_id, auth):
    servers = Server.get_neighbours(exclude=server_id)
    tasks = []

    # route information
    tasks.append(routing.async_remove_neighbour_send(server_id, auth=auth, servers=servers))
    await asyncio.gather(*tasks)


@api_bp.route('/cluster/out/<server_id>', methods=['POST'])
@jwt_required
@securizer
def cluster_out(server_id):
    user = User.get_current()
    if user and user.user == 'root':
        Server.query.get_or_raise(server_id)
        data = request.get_json()
        if data.get('death', None):
            try:
                death = dt.datetime.strptime(data['death'], defaults.DATEMARK_FORMAT)
            except:
                death = get_now()
        else:
            death = None
        current_app.cluster_manager.set_death(server_id, death=death)
        _cluster_logger.debug(f"{Server.query.get(server_id).name or server_id} is a death server")

        # run background route
        executor.submit(asyncio.run, background_cluster_out(server_id, get_auth_from_request()))

        return {}, 204
    else:
        raise errors.UserForbiddenError


# @api_bp.route('/routes/<server_id>', methods=['GET'])
# @forward_or_dispatch()
# @jwt_required
# @securizer
# def routes_neighbour(server_id):
#     server = Server.query.get_or_raise(server_id)
#     route = routing.check_neighbour(server)
#     # change server
#     if route:
#         return {'neighbour': True}
#     else:
#         return {'neighbour': False}


# set node to alive and check if neighbour

@api_bp.route('/routes', methods=['GET', 'POST', 'PATCH'])
@log_time('full')
@forward_or_dispatch('GET', 'POST')
@jwt_required
@securizer
@validate_schema(POST=routes_post, PATCH=routes_patch)
@log_time('after validation')
def routes():
    if request.method == 'GET':
        route_table = []
        for route in Route.query.join(Server.route).order_by(
                Server.name).filter(Server.deleted == False).all():
            route_table.append(route.to_json(human=check_param_in_uri('human')))
        data = {'route_list': route_table}
        data.update(server=dict(name=g.server.name, id=str(g.server.id)))
        return data

    elif request.method == 'POST':
        async def job(msg, auth):
            new_routes = await routing.async_update_route_table_cost(**msg)
            if len(new_routes) > 0:
                await routing.async_send_routes(auth=auth)
            return new_routes

        msg = request.get_json()
        if msg.get('background', False):
            executor.submit(asyncio.run, job(msg, get_auth_from_request()))
        else:
            asyncio.run(job(msg, HTTPBearerAuth(request.headers['Authorization'].split()[1])))
        return {}, 204

    elif request.method == 'PATCH':
        def job(msg, auth):
            new_routes = routing.update_route_table_from_data(msg, auth)
            if len(new_routes) > 0:
                exclude = msg.get('exclude', [])
                exclude.append(msg.get('server_id'))
                asyncio.run(routing.async_send_routes(new_routes, auth, exclude=exclude))

        executor.submit(job, request.get_json(), HTTPBearerAuth(request.headers['Authorization'].split()[1]))

        return {}, 204


def run_command_and_callback(operation: 'IOperationEncapsulation', var_context, source: Server,
                             step_execution: StepExecution,
                             jwt_identity, event_id,
                             timeout=None):
    cp = operation.execute(var_context=var_context, timeout=timeout)

    execution = db.session.merge(step_execution)
    source = db.session.merge(source)
    execution.load_completed_result(cp)
    try:
        db.session.commit()
    except Exception as e:
        current_app.logger.exception(f"Error on commit for execution {execution.id}")
    data = dict(step_execution=execution.to_json(),
                var_context=base64.b64encode(pickle.dumps(var_context)).decode('ascii'))
    if execution.child_orch_execution:
        data['step_execution'].update(orch_execution=execution.child_orch_execution.to_json(add_step_exec=True))
    resp, code = ntwrk.post(server=source, view_or_url='api_1_0.events', view_data={'event_id': event_id},
                            json=data,
                            auth=HTTPBearerAuth(
                                create_access_token(jwt_identity, expires_delta=dt.timedelta(seconds=30))))
    if code != 202:
        current_app.logger.error(f"Error while sending result for execution {execution.id}: {code}, {resp}")
    return data


@api_bp.route('/launch/operation', methods=['POST'])
@forward_or_dispatch()
@jwt_required
@securizer
def launch_operation():
    data = request.get_json()
    operation = pickle.loads(base64.b64decode(data['operation'].encode('ascii')))
    var_context = pickle.loads(base64.b64decode(data['var_context'].encode('ascii')))
    orch_exec_json = data['orch_execution']
    orch_exec = OrchExecution.from_json(orch_exec_json)

    se = StepExecution(id=var_context.globals.get('step_execution_id'), server=g.server, step_id=data.get('step_id'),
                       orch_execution=orch_exec, params=dict(var_context),
                       start_time=get_now())
    db.session.add_all([orch_exec, se])
    db.session.commit()
    future = executor.submit(run_command_and_callback, operation, var_context, g.source, se,
                             get_jwt_identity(), data['event_id'],
                             timeout=data.get('timeout', None))
    try:
        r = future.result(1)
    except concurrent.futures.TimeoutError:
        return {}, 204
    except Exception as e:
        current_app.logger.exception(
            f"Exception got when executing step {data.get('step_id')}. See logs for more information")
        return {'error': f"Exception got when executing step {data.get('step_id')}. See logs for more information"}, 500
    else:
        return r, 200


@api_bp.route('/launch/orchestration/<orchestration_id>', methods=['POST'])
@forward_or_dispatch()
@jwt_required
@securizer
@validate_schema(launch_orchestration_post)
def launch_orchestration(orchestration_id):
    orchestration = Orchestration.query.get_or_raise(orchestration_id)
    params = request.get_json().get('params', {})
    hosts = request.get_json().get('hosts')

    a = set(orchestration.target)
    if not isinstance(hosts, dict):
        hosts = dict(all=hosts)
    b = set(hosts.keys())
    c = a - b
    if len(c) > 0:
        raise errors.TargetUnspecified(c)
    c = b - a
    if len(c) > 0:
        raise errors.TargetNotNeeded(c)

    not_found = normalize_hosts(hosts)
    if not_found:
        raise errors.ServerNormalizationError(not_found)

    for target, target_hosts in hosts.items():
        if len(target_hosts) == 0:
            raise errors.EmptyTarget(target)
    # check param entries
    # rest = orchestration.user_parameters - set(params.keys())
    # if rest:
    #     rest = list(rest)
    #     rest.sort()
    #     return {'error': f"Parameter(s) not specified: {', '.join(rest)}"}, 404

    if not orchestration.steps:
        return errors.GenericError('orchestration does not have steps to execute', orchestration_id=orchestration_id)

    execution_id = str(uuid.uuid4())

    vc = VarContext(globals=dict(orch_execution_id=execution_id, executor_id=get_jwt_identity()), initials=params)

    if request.get_json().get('background', True):
        future = executor.submit(deploy_orchestration, orchestration=orchestration.id, var_context=vc, hosts=hosts,
                                 execution=execution_id, )
        try:
            r = future.result(1)
        except concurrent.futures.TimeoutError:
            return {'execution_id': execution_id}, 202
        except Exception as e:
            current_app.logger.exception(f"Exception got when executing orchestration {orchestration}")
            raise
        else:
            return jsonify(r), 200
    else:
        try:
            orch_exe = deploy_orchestration(orchestration=orchestration, var_context=vc, hosts=hosts,
                                            execution=execution_id)
        except Exception as e:
            current_app.logger.exception(f"Exception got when executing orchestration {orchestration}")
            raise
        else:
            return orch_exe.to_json(add_step_exec=True, human=check_param_in_uri('human'), split_lines=True), 200


def wrap_sudo(user, cmd):
    if isinstance(user, User):
        username = user.name
    else:
        username = user
    if isinstance(cmd, str):
        return f"sudo -Siu {username} -- sh -c {shlex.quote(cmd)}"
    else:
        args = ["sudo", "-Siu", username, "--", "bash" "-c",  shlex.quote(cmd)]
        return args


@api_bp.route('/launch/command', methods=['POST'])
@forward_or_dispatch()
@jwt_required
@securizer
@validate_schema(launch_command_post)
def launch_command():
    data = request.get_json()

    server_list = []
    if 'target' in data:
        not_found = []
        servers = Server.query.all()
        if data['target'] == 'all':
            server_list = servers
        elif is_iterable_not_string(data['target']):
            for vv in data['target']:
                sl = search(vv, servers)
                if len(sl) == 0:
                    not_found.append(vv)
                else:
                    server_list.extend(sl)
        else:
            sl = search(data['target'], servers)
            if len(sl) == 0:
                not_found.append(data['target'])
            else:
                server_list.extend(sl)
        if not_found:
            return {'error': "Following granules or ids did not match to any server: " + ', '.join(not_found)}, 404
    else:
        server_list.append(g.server)

    if re.search(r'rm\s+((-\w+|--[-=\w]*)\s+)*(-\w*[rR]\w*|--recursive)', data['command']):
        return {'error': 'rm with recursion is not allowed'}, 403
    data.pop('target', None)
    start = None

    username = getattr(User.query.get(get_jwt_identity()), 'user', None)
    if not username:
        raise errors.EntityNotFound('User', get_jwt_identity())
    cmd = wrap_sudo(username, data['command'])
    if g.server in server_list:
        start = time.time()
        server_list.pop(server_list.index(g.server))
        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE if data.get('input', None) else None, stderr=subprocess.PIPE,
                                stdout=subprocess.PIPE, shell=True, close_fds=True, encoding='utf-8')

    resp_data = {}
    if check_param_in_uri("human"):
        attr = 'name'
    else:
        attr = 'id'
    if server_list:
        resp: t.List[ntwrk.Response] = asyncio.run(
            ntwrk.parallel_requests(server_list, method='POST', view_or_url='api_1_0.launch_command', json=data,
                                    auth=get_auth_from_request()))
        for s, r in zip(server_list, resp):
            key = getattr(s, attr, s.id)
            if r.ok:
                resp_data[key] = r.msg[s.id]
            else:
                if not r.exception:
                    resp_data[key] = {
                        'error': {'status_code': r.code, 'response': r.msg}}
                else:
                    if isinstance(r.exception, errors.BaseError):
                        resp_data[key] = errors.format_error_content(r.exception, current_app.config['DEBUG'])
                    else:
                        resp_data[key] = {
                            'error': format_exception(r.exception) if current_app.config['DEBUG'] else str(
                                r.exception) or str(r.exception.__class__.__name__)}

    if start:
        key = getattr(g.server, attr, g.server.id)
        timeout = data.get('timeout', defaults.TIMEOUT_COMMAND)
        try:
            outs, errs = proc.communicate(input=(data.get('input', '') or ''), timeout=timeout)
        except (TimeoutError, subprocess.TimeoutExpired):
            proc.kill()
            try:
                outs, errs = proc.communicate(timeout=1)
            except:
                resp_data[key] = {
                    'error': f"Command '{cmd}' timed out after {timeout} seconds. Unable to communicate with the process launched."}
            else:
                resp_data[key] = {
                    'error': f"Command '{cmd}' timed out after {timeout} seconds",
                    'stdout': outs.split('\n'), 'stderr': errs.split('\n')}
        except Exception as e:
            current_app.logger.exception("Exception raised while trying to run command")
            resp_data[key] = {
                'error': traceback.format_exc() if current_app.config['DEBUG'] else str(
                    r.exception) or str(r.exception.__class__.__name__)}
        else:
            resp_data[key] = {'stdout': outs.split('\n'), 'stderr': errs.split('\n'),
                                           'returncode': proc.returncode}
    resp_data['cmd'] = cmd
    resp_data['input'] = data.get('input', None)
    return resp_data, 200


@api_bp.route('/events/<event_id>', methods=['POST'])
@forward_or_dispatch()
@jwt_required
@securizer
def events(event_id):
    e = Event(event_id, data=request.get_json())
    current_app.events.dispatch(e)
    return {}, 202


@api_bp.route('/orchestrations/full', methods=['POST'])
@forward_or_dispatch()
@jwt_required
@securizer
@validate_schema(orchestration_full)
@lock_catalog
def orchestrations_full():
    json_data = request.get_json()
    json_steps = json_data.pop('steps')
    generated_version = False
    if 'version' not in json_data:
        generated_version = True
        json_data['version'] = Orchestration.query.filter_by(name=json_data['name']).count() + 1
    o = Orchestration(**json_data)
    db.session.add(o)
    resp_data = {'id': str(o.id)}
    if generated_version:
        resp_data.update(version=o.version)

    # reorder steps in order of dependency
    id2step = {str(s['id']): s for s in json_steps}

    dag = DAG()
    for s in json_steps:
        step_id = str(s['id'])
        if s['undo'] and len(s.get('parent_step_ids', [])) == 0:
            raise errors.UndoStepWithoutParent(step_id)
        dag.add_node(step_id)
        for p_s_id in s.get('parent_step_ids', []):
            dag.add_edge(str(p_s_id), step_id)

    if dag.is_cyclic():
        raise errors.CycleError

    new_steps = []
    for step_id in dag.ordered_nodes:
        step = id2step[step_id]
        new_steps.append(step)
    # end reorder steps in order of dependency

    rid2step = OrderedDict()
    dependencies = {}
    for json_step in new_steps:
        rid = str(json_step.pop('id', None))
        if rid is not None and rid in rid2step.keys():
            raise errors.DuplicatedId(rid)
        if 'action_template_id' in json_step:
            json_step['action_template'] = ActionTemplate.query.get_or_raise(json_step.pop('action_template_id'))
        elif 'action_type' in json_step:
            json_step['action_type'] = ActionType[json_step.pop('action_type')]
        dependencies[rid] = {'parent_step_ids': [str(p_id) for p_id in json_step.pop('parent_step_ids', [])]}
        s = o.add_step(**json_step)
        db.session.add(s)
        if rid:
            rid2step[rid] = s

        continue

    # process dependencies
    for rid, dep in dependencies.items():
        step = rid2step[rid]
        parents = []
        for p_s_id in dep['parent_step_ids']:
            if p_s_id in rid2step:
                parents.append(rid2step[p_s_id])
        o.set_parents(step, parents)

    db.session.commit()

    # send new ids in order of appearance at beginning
    new_id_steps = []
    for rid in rid2step.keys():
        new_id_steps.append(str(rid2step[rid].id))
    resp_data.update({'step_ids': new_id_steps})
    return resp_data, 201



import json
import logging
from pprint import pprint

import pygments
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import PygmentsTokens
from pygments.lexers.web import JSONLexer

from dimensigon.dshell import environ as env
from dimensigon.utils.helpers import format_exception
from dimensigon.web.network import Response

logger = logging.getLogger('dshell')

def dprint(msg):
    if isinstance(msg, str):
        print(msg)
    elif isinstance(msg, Response):
        if msg.code:
            dprint(msg.msg)
        else:
            dprint(msg.exception)
    elif isinstance(msg, Exception):
        if env.get('DEBUG'):
            dprint(format_exception(msg))
        else:
            dprint(str(msg) if str(msg) else msg.__class__.__name__)
    else:
        try:
            tokens = list(pygments.lex(json.dumps(msg, indent=2), lexer=JSONLexer()))
        except:
            pprint(msg)
        else:
            print_formatted_text(PygmentsTokens(tokens), end="")



from contextlib import contextmanager

from flask import has_app_context, current_app

from dimensigon.domain.entities import Server, Locker, User, ActionTemplate
from dimensigon.web import db


@contextmanager
def prepare_context(app=None):
    ctx = None
    if app:
        ctx = app.app_context()
        ctx.push()

    if not has_app_context():
        ctx = current_app.app_context()
        ctx.push()

    yield ctx

    if ctx:
        ctx.pop()


def _bootstrap_database():
    db.create_all()


def set_initial(app=None, session=None, server=True, user=True, action_template=False):
    """Used for generate database on testing"""
    with prepare_context(app) as ctx:
        _bootstrap_database()
        Locker.set_initial()
        if server:
            Server.set_initial()
        if user:
            User.set_initial()
        if action_template:
            ActionTemplate.set_initial()

        if ctx:
            db.session.commit()




import datetime as dt
import logging
import logging.config
import multiprocessing
import os
import sys
from copy import deepcopy

from dimensigon import defaults
from dimensigon.__main__ import RuntimeConfig, Dimension
from dimensigon.core import Config
from dimensigon.core import Dimensigon
from dimensigon.db import setup_db
from dimensigon.web import config_by_name


def get_default_config_dir() -> object:
    """Put together the default configuration directory based on the OS."""
    data_dir = os.getenv("APPDATA") if os.name == "nt" else os.path.expanduser("~")
    return os.path.join(data_dir, defaults.CONFIG_DIR_NAME)  # type: ignore


def ensure_config_path(config_dir: str) -> None:
    """Validate the configuration directory."""

    # Test if configuration directory exists
    if not os.path.isdir(config_dir):
        if config_dir != get_default_config_dir():
            print(
                f"Fatal Error: Specified configuration directory {config_dir} "
                "does not exist"
            )
            sys.exit(1)

        try:
            os.mkdir(config_dir)
        except OSError as e:
            print(
                "Fatal Error: Unable to create default configuration "
                f"directory {config_dir}. Error: {e}"
            )
            sys.exit(1)

        _write_default_config(config_dir)


default_logconfig_dict = {
    'version': 1,
    'disable_existing_loggers': False,
    "root": {"level": "INFO", "handlers": ["console", "error_file"]},
    'loggers': {
        "gunicorn.error": {
            "level": "INFO",
            "propagate": True,
            "qualname": "gunicorn.error"
        },
        "gunicorn.access": {
            "level": "INFO",
            "handlers": ["access_file"],
            "propagate": False,
            "qualname": "gunicorn.access"
        },
        "urllib3": {
            "level": "ERROR",
            "propagate": False,
        },
        # "dimensigon": {
        #     "level": "DEBUG",
        #     "qualname": "dimensigon"
        # },
        # "dimensigon.db": {
        #     "level": "DEBUG",
        #     "qualname": "dimensigon.dm"
        # },
        # "dimensigon.routing": {
        #     "level": "DEBUG",
        #     "qualname": "dimensigon.routing"
        # },
        # "dimensigon.cluster": {
        #     "level": "DEBUG",
        #     "qualname": "dimensigon.cluster"
        # },
        # "dimensigon.catalog": {
        #     "level": "DEBUG",
        #     "qualname": "dimensigon.catalog"
        # },
        # "dimensigon.network": {
        #     "level": "INFO",
        #     "qualname": "dimensigon.network"
        # },
        # "dimensigon.query": {
        #     "level": "DEBUG",
        #     "handlers": ["query_file"],
        #     "propagate": False
        # },
        "apscheduler": {
            "level": "ERROR",
            "qualname": "apscheduler"
        },
        # "apscheduler.scheduler": {
        #     "level": "INFO",
        #     "qualname": "apscheduler.scheduler"
        # },
        # "apscheduler.executors": {
        #     "level": "INFO",
        #     "qualname": "apscheduler.executors.default"
        # },
        # "sqlalchemy.engine": {
        #     "level": "ERROR",
        # }
        "asyncio": {
            "propagate": False
        },
        # "sqlalchemy.engine": {
        #     "level": "INFO"
        # }
    },
    'handlers': {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "generic",
            "stream": "ext://sys.stdout"
        },
        "error_file": {
            "level": "INFO",
            "class": "logging.handlers.RotatingFileHandler",
            "maxBytes": 20 * 1024 * 1024,
            "backupCount": 5,
            "formatter": "detailed",
            "filename": "dimensigon.log"
        },
        "query_file": {
            "class": "logging.handlers.RotatingFileHandler",
            "maxBytes": 20 * 1024 * 1024,
            "backupCount": 5,
            "formatter": "detailed",
            "filename": "query.log"
        },
        "access_file": {
            "class": "logging.handlers.RotatingFileHandler",
            "maxBytes": 20 * 1024 * 1024,
            "backupCount": 5,
            "formatter": "access",
            "filename": "access.log",
        }
    },
    'formatters': {
        "generic": {
            "format": "%(asctime)s %(levelname)-8s %(name)-24s %(message)s",
            "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
            "class": "logging.Formatter"
        },
        "detailed": {
            "format": "%(asctime)s [%(thread)-8s] %(levelname)-8s %(name)-30s %(message)s",
            "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
            "class": "logging.Formatter"
        },
        "access": {
            "format": "%(message)s",
            "class": "logging.Formatter"
        }
    }
}

logconfig = {}


def _setup_dimensigon_config(run_config: RuntimeConfig, config: Config):
    global logconfig
    # save arguments passed
    config.args = sys.argv

    config.config_dir = run_config.config_dir or get_default_config_dir()
    ensure_config_path(config.config_dir)

    config.debug = run_config.debug
    config.flask = run_config.flask
    config.refresh_interval = dt.timedelta(minutes=run_config.refresh_interval)
    config.force_scan = run_config.force_scan

    logconfig = deepcopy(default_logconfig_dict)
    for k, v in run_config.logconfig.items():
        if k in logconfig:
            logconfig[k].update(v)
        else:
            logconfig[k] = v

    for name, handler in logconfig['handlers'].items():
        if name == 'error_file':
            handler['filename'] = run_config.errorlog or config.path(defaults.LOG_FOLDER, defaults.ERROR_LOGFILE)
        elif name == 'access_file':
            handler['filename'] = run_config.accesslog or config.path(defaults.LOG_FOLDER, defaults.ACCESS_LOGFILE)
        else:
            if 'filename' in handler and not os.path.dirname(handler['filename']):
                handler['filename'] = config.path(defaults.LOG_FOLDER, handler['filename'])

    logconfig['root']['level'] = 'DEBUG' if run_config.debug else 'INFO'

    # in case log gunicorn not called
    logging.config.dictConfig(logconfig)


def setup_database_uri(run_config: RuntimeConfig, config: Config):
    db_path = os.path.join(config.config_dir, defaults.DEFAULT_DB_FILE)
    config.db_uri = defaults.DEFAULT_DB_URL.format(db_file=db_path)


def _setup_http_config(run_config: RuntimeConfig, config: Config):
    def on_exit(server):
        server.app.dm.flask_app.shutdown()

    bind = []
    for ip in run_config.ips or ['0.0.0.0']:
        bind.append(f"{ip}:{run_config.port or defaults.DEFAULT_PORT}")

    config.http_conf.update(  # Logging
        access_log_format='%(h)s %(l)s %(u)s %(t)s "%(r)s" %(s)s %(b)s %(L)s "%(f)s" "%(a)s"',
        capture_output=False,
        logconfig_dict=logconfig,
        enable_stdio_inheritance=True,
        # Process Naming
        proc_name=defaults.PROC_NAME,
        # SSL
        # do_handshake_on_connect=False,
        # Server Hooks
        on_exit=on_exit,
        # Server Mechanics
        preload_app=True,
        daemon=run_config.daemon,
        pidfile=os.path.join(run_config.pid_file or config.config_dir, defaults.PID_FILE),
        # Server Socket
        bind=bind,
        # Worker Processes
        workers=1,
        worker_class='gthread',
        threads=run_config.threads or max(12, 4 * multiprocessing.cpu_count()),
        # threads=4,
        # max_requests=100,
        timeout=300,
        graceful_timeout=60,
    )

    if run_config.certfile:
        if not os.path.exists(run_config.certfile):
            raise FileNotFoundError(run_config.certfile)
        else:
            config.http_conf.update(certfile=run_config.certfile)
    else:
        cert_file = os.path.join(config.config_dir, defaults.SSL_DIR, defaults.CERT_FILE)
        config.http_conf.update(certfile=cert_file)
    if run_config.keyfile:
        if not os.path.exists(run_config.keyfile):
            raise FileNotFoundError(run_config.keyfile)
        else:
            config.http_conf.update(keyfile=run_config.keyfile)
    else:
        key_file = os.path.join(config.config_dir, defaults.SSL_DIR, defaults.KEY_FILE)
        config.http_conf.update(keyfile=key_file)


def _setup_flask_config(run_config: RuntimeConfig, dm: Dimensigon):
    flask_config = config_by_name['gunicorn']
    flask_config.SQLALCHEMY_DATABASE_URI = dm.config.db_uri
    if run_config.debug:
        flask_config.DEBUG = run_config.debug

    dm.config.flask_conf = flask_config

    result = dm.engine.execute(Dimension.__table__.select(Dimension.current).where(Dimension.current == True))
    result = result.fetchall()
    if len(result) == 1:
        dm.config.flask_conf.SECRET_KEY = result[0][0]
    elif len(result) > 1:
        raise ValueError('More than one dimension are set to current.')
    else:
        dm.config.flask_conf.SECRET_KEY = 'my_precious_key'


def setup_dm(run_config: RuntimeConfig) -> Dimensigon:
    dm = Dimensigon()

    # set dimensigon configuration
    _setup_dimensigon_config(run_config, dm.config)

    # set http configuration. Before setup_db to get ip binds
    _setup_http_config(run_config, dm.config)

    # set database uri
    setup_database_uri(run_config, dm.config)

    # set database to allow queries
    setup_db(dm)

    # set flask configuration
    _setup_flask_config(run_config, dm)

    return dm


def _write_default_config(config_dir: str) -> bool:
    """Write the default config."""

    software_repo_path = os.path.join(config_dir, defaults.SOFTWARE_REPO)
    log_folder_path = os.path.join(config_dir, defaults.LOG_FOLDER)
    log_sender_repo_path = os.path.join(config_dir, defaults.LOG_SENDER_REPO)
    ssl_path = os.path.join(config_dir, defaults.SSL_DIR)

    try:
        os.makedirs(software_repo_path, exist_ok=True)
        os.makedirs(log_folder_path, exist_ok=True)
        os.makedirs(log_sender_repo_path, exist_ok=True)
        os.makedirs(ssl_path, exist_ok=True)
    except OSError:
        print("Unable to create default configuration", config_dir)
        return False



import datetime as dt
from unittest import TestCase, mock

from dimensigon.utils.cluster import ClusterManager, _ClusterRegister, ClusterManagerCoordinator, \
    date_format
from dimensigon.utils.helpers import get_now

now = get_now()
str_now = now.strftime(date_format)

before_now = now - dt.timedelta(seconds=1)
after_now = now + dt.timedelta(seconds=1)


class TestClusterManager(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.cm = ClusterManager()
        cls.maxDiff = None

    def setUp(self) -> None:
        self.cm.clear_cluster()

    @mock.patch('dimensigon.utils.cluster_manager.get_now')
    def test_set_alive(self, mock_now):
        mock_now.return_value = now

        self.assertIsNotNone(self.cm.set_alive(1))
        self.assertDictEqual({1: _ClusterRegister(1, birth=now, keepalive=now, death=None)}, self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, death=before_now)
        self.assertIsNotNone(self.cm.set_alive(1))
        self.assertDictEqual({1: _ClusterRegister(1, birth=now, keepalive=now, death=None)}, self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, death=None)
        self.assertIsNone(self.cm.set_alive(1))
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, death=None)}, self.cm._cluster)

        self.assertIsNotNone(self.cm.set_alive(2))
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, death=None),
                              2: _ClusterRegister(2, birth=now, keepalive=now, death=None)}, self.cm._cluster)

    @mock.patch('dimensigon.utils.cluster_manager.get_now')
    def test_set_death(self, mock_now):
        mock_now.return_value = now

        self.assertIsNotNone(self.cm.set_death(1))
        self.assertDictEqual({1: _ClusterRegister(1, birth=now, death=now)}, self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, keepalive=before_now, death=before_now)
        self.assertIsNone(self.cm.set_death(1))
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, keepalive=before_now, death=before_now)},
                             self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, keepalive=before_now, death=None)
        self.assertIsNotNone(self.cm.set_death(1))
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, keepalive=before_now, death=now)},
                             self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=after_now, keepalive=after_now, death=None)
        self.assertIsNotNone(self.cm.set_death(1))
        self.assertDictEqual({1: _ClusterRegister(1, birth=after_now, keepalive=after_now, death=after_now)},
                             self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=after_now, keepalive=after_now, death=None)
        self.assertIsNotNone(self.cm.set_death(1, before_now))
        self.assertDictEqual({1: _ClusterRegister(1, birth=after_now, keepalive=after_now, death=after_now)},
                             self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, keepalive=before_now, death=None)
        self.assertIsNotNone(self.cm.set_death(1, after_now))
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, keepalive=before_now, death=after_now)},
                             self.cm._cluster)

    @mock.patch('dimensigon.utils.cluster_manager.get_now')
    def test_set_keepalive(self, mock_now):
        mock_now.return_value = now

        self.cm.set_keepalive(1)
        self.assertDictEqual({1: _ClusterRegister(1, birth=now, keepalive=now, death=None)},
                             self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, death=before_now)
        self.cm.set_keepalive(1)
        self.assertDictEqual({1: _ClusterRegister(1, birth=now, keepalive=now, death=None)},
                             self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, death=None)
        self.cm.set_keepalive(1)
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, keepalive=now, death=None)}, self.cm._cluster)

        self.cm.set_keepalive(2)
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, keepalive=now, death=None),
                              2: _ClusterRegister(2, birth=now, keepalive=now, death=None)},
                             self.cm._cluster)

        self.cm.threshold = dt.timedelta(seconds=1)
        self.cm.set_keepalive(2, after_now)
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, keepalive=now, death=None),
                              2: _ClusterRegister(2, birth=now, keepalive=now, death=None)},
                             self.cm._cluster)

    def test__update_cluster(self):
        self.cm._cluster = {1: _ClusterRegister(1, birth=before_now, keepalive=before_now, death=None),
                            2: _ClusterRegister(2, birth=now, keepalive=now, death=None),
                            3: _ClusterRegister(3, birth=before_now, keepalive=before_now, death=None),
                            4: _ClusterRegister(4, birth=before_now, keepalive=before_now, death=now),
                            5: _ClusterRegister(5, birth=before_now, keepalive=before_now, death=now),
                            6: _ClusterRegister(6, birth=before_now, keepalive=before_now, death=after_now),
                            7: _ClusterRegister(7, birth=before_now, keepalive=before_now, death=now),
                            8: _ClusterRegister(8, birth=before_now, keepalive=before_now, death=now),
                            10: _ClusterRegister(10, birth=before_now, keepalive=before_now, death=None)}

        self.cm._update_cluster([_ClusterRegister(1, birth=now, keepalive=now, death=None),
                                 _ClusterRegister(2, birth=before_now, keepalive=before_now, death=None),
                                 _ClusterRegister(3, birth=before_now, keepalive=before_now, death=now),
                                 _ClusterRegister(4, birth=before_now, keepalive=before_now, death=None),
                                 _ClusterRegister(5, birth=before_now, keepalive=before_now, death=after_now),
                                 _ClusterRegister(6, birth=before_now, keepalive=before_now, death=now),
                                 _ClusterRegister(7, birth=after_now, keepalive=after_now, death=None),
                                 _ClusterRegister(8, birth=after_now, keepalive=after_now, death=after_now),
                                 _ClusterRegister(9, birth=before_now, keepalive=before_now, death=None)])

        self.assertDictEqual({1: _ClusterRegister(1, birth=now, keepalive=now, death=None),
                          2: _ClusterRegister(2, birth=now, keepalive=now, death=None),
                          3: _ClusterRegister(3, birth=before_now, keepalive=before_now, death=now),
                          4: _ClusterRegister(4, birth=before_now, keepalive=before_now, death=now),
                          5: _ClusterRegister(5, birth=before_now, keepalive=before_now, death=after_now),
                          6: _ClusterRegister(6, birth=before_now, keepalive=before_now, death=after_now),
                          7: _ClusterRegister(7, birth=after_now, keepalive=after_now, death=None),
                          8: _ClusterRegister(8, birth=after_now, keepalive=after_now, death=after_now),
                          9: _ClusterRegister(9, birth=before_now, keepalive=before_now, death=None),
                          10: _ClusterRegister(10, birth=before_now, keepalive=before_now, death=None)},
                         self.cm._cluster)

    def test_get_cluster(self):
        self.cm._cluster[1] = _ClusterRegister(1, birth=now, death=None)

        self.assertListEqual([dict(id=1, birth=str_now)], self.cm.get_cluster())

    def test_contains(self):
        self.cm._cluster[1] = _ClusterRegister(1, birth=now, death=None)
        self.cm._cluster[2] = _ClusterRegister(2, birth=now, death=now)

        self.assertTrue(1 in self.cm)
        self.assertTrue(2 not in self.cm)

    def test_iter(self):
        self.cm._cluster[1] = _ClusterRegister(1, birth=now, death=None)
        self.cm._cluster[2] = _ClusterRegister(2, birth=now, death=now)

        self.assertListEqual([1], [i for i in self.cm])

    def test_get_oldest_alive(self):
        self.cm._cluster = {1: _ClusterRegister(1, birth=now, death=None),
                            2: _ClusterRegister(2, birth=before_now, death=None)}

        self.assertEqual(2, self.cm.get_oldest_alive())

    @mock.patch('dimensigon.utils.cluster_manager.get_now')
    def test_get_delta_keepalive(self, mock_now):
        mock_now.return_value = after_now
        self.cm._cluster = {1: _ClusterRegister(1, birth=now, keepalive=now, death=None),
                            2: _ClusterRegister(2, birth=after_now, keepalive=after_now, death=None)}

        self.assertListEqual([], self.cm.get_delta_keepalive(dt.timedelta(seconds=0)))
        self.assertListEqual([2], self.cm.get_delta_keepalive(dt.timedelta(seconds=1)))
        self.assertListEqual([1, 2], self.cm.get_delta_keepalive(dt.timedelta(seconds=2)))


class TestClusterManagerCoordinator(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.cm = ClusterManagerCoordinator()
        cls.maxDiff = None

    def setUp(self) -> None:
        self.cm.clear_cluster()

    @mock.patch('dimensigon.utils.cluster_manager.get_now')
    def test_set_alive(self, mock_now):
        mock_now.return_value = now

        self.cm.set_alive(1)
        self.assertDictEqual({1: _ClusterRegister(1, birth=now, keepalive=now, death=None, coordinator=True)}, self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, death=before_now)
        self.cm.set_alive(1)
        self.assertDictEqual({1: _ClusterRegister(1, birth=now, keepalive=now, death=None, coordinator=True)}, self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, death=None)
        self.cm.set_alive(1)
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, death=None, coordinator=True)}, self.cm._cluster)

        self.cm.set_alive(2)
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, death=None, coordinator=True),
                              2: _ClusterRegister(2, birth=now, keepalive=now, death=None, coordinator=False)}, self.cm._cluster)

    @mock.patch('dimensigon.utils.cluster_manager.get_now')
    def test_set_death(self, mock_now):
        mock_now.return_value = now

        self.cm.set_death(1)
        self.assertDictEqual({1: _ClusterRegister(1, birth=now, death=now, coordinator=False)}, self.cm._cluster)

        self.cm._cluster[1] = _ClusterRegister(1, birth=before_now, death=None, coordinator=True)
        self.cm._cluster[2] = _ClusterRegister(2, birth=now, death=None, coordinator=False)
        self.cm._cluster[3] = _ClusterRegister(3, birth=before_now, death=None, coordinator=False)
        self.cm.set_death(1)
        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, death=now, coordinator=False),
                              2: _ClusterRegister(2, birth=now, death=None, coordinator=False),
                              3: _ClusterRegister(3, birth=before_now, death=None, coordinator=True)}, self.cm._cluster)


    def test__update_cluster(self):
        self.cm._cluster = {1: _ClusterRegister(1, birth=before_now, death=None, coordinator=True),
                            2: _ClusterRegister(2, birth=now, death=None, coordinator=True),
                            3: _ClusterRegister(3, birth=before_now, death=None),
                            4: _ClusterRegister(4, birth=before_now, death=now),
                            5: _ClusterRegister(5, birth=before_now, death=now),
                            6: _ClusterRegister(6, birth=before_now, death=after_now),
                            7: _ClusterRegister(7, birth=before_now, death=now),
                            8: _ClusterRegister(8, birth=before_now, death=now),
                            10: _ClusterRegister(10, birth=before_now, death=None)}

        self.cm._update_cluster([_ClusterRegister(1, birth=now, death=None),
                                 _ClusterRegister(2, birth=before_now, death=None),
                                 _ClusterRegister(3, birth=before_now, death=now),
                                 _ClusterRegister(4, birth=before_now, death=None),
                                 _ClusterRegister(5, birth=before_now, death=after_now),
                                 _ClusterRegister(6, birth=before_now, death=now),
                                 _ClusterRegister(7, birth=after_now, death=None, coordinator=True),
                                 _ClusterRegister(8, birth=after_now, death=after_now),
                                 _ClusterRegister(9, birth=before_now, death=None)])

        self.assertDictEqual({1: _ClusterRegister(1, birth=before_now, death=None, coordinator=True),
                              2: _ClusterRegister(2, birth=before_now, death=None),
                              3: _ClusterRegister(3, birth=before_now, death=now),
                              4: _ClusterRegister(4, birth=before_now, death=now),
                              5: _ClusterRegister(5, birth=before_now, death=now),
                              6: _ClusterRegister(6, birth=before_now, death=now),
                              7: _ClusterRegister(7, birth=after_now, death=None),
                              8: _ClusterRegister(8, birth=after_now, death=after_now),
                              9: _ClusterRegister(9, birth=before_now, death=None),
                              10: _ClusterRegister(10, birth=before_now, death=None)}, self.cm._cluster)

    def test_coordinators(self):
        self.cm._cluster = {1: _ClusterRegister(1, birth=now, death=None, coordinator=False),
                            2: _ClusterRegister(2, birth=before_now, death=None, coordinator=True),
                            3: _ClusterRegister(3, birth=before_now, death=None, coordinator=True)}

        self.assertListEqual([_ClusterRegister(2, birth=before_now, death=None, coordinator=True),
                              _ClusterRegister(3, birth=before_now, death=None, coordinator=True)],
                             self.cm.coordinators)

    def test_get_older_coordinators(self):
        self.cm._cluster = {1: _ClusterRegister(1, birth=now, death=None, coordinator=False),
                            2: _ClusterRegister(2, birth=after_now, death=None, coordinator=True),
                            3: _ClusterRegister(3, birth=before_now, death=None, coordinator=True)}

        self.assertEqual(3, self.cm.get_oldest_coordinator())

    def test_set_coordinator(self):
        self.cm._cluster = {1: _ClusterRegister(1, birth=now, death=None, coordinator=False),
                            2: _ClusterRegister(2, birth=after_now, death=None, coordinator=True),
                            3: _ClusterRegister(3, birth=before_now, death=None, coordinator=True)}

        self.cm.set_coordinator(1)

        self.assertDictEqual({1: _ClusterRegister(1, birth=now, death=None, coordinator=True),
                              2: _ClusterRegister(2, birth=after_now, death=None, coordinator=False),
                              3: _ClusterRegister(3, birth=before_now, death=None, coordinator=False)},
                             self.cm._cluster)


# class TestClusterManagerSession(TestCase):
#
#     @classmethod
#     def setUpClass(cls) -> None:
#         cls.cm = ClusterManagerSession()
#         cls.maxDiff = None
#
#     def setUp(self) -> None:
#         self.cm.clear_cluster()
#
#     @mock.patch('dimensigon.utils.cluster_manager.get_now')
#     def test_set_alive(self, mock_now):
#         mock_now.return_value = now
#
#         self.assertIsNotNone(self.cm.set_alive(1, 11))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=now, keepalive=now, death=None)}, self.cm._cluster)
#
#         self.cm._cluster[1] = _ClusterRegisterSession(1, session=11, birth=before_now, death=before_now)
#         self.assertIsNotNone(self.cm.set_alive(1, 11))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=now, death=None)}, self.cm._cluster)
#
#         self.cm._cluster[1] = _ClusterRegisterSession(1, session=11, birth=before_now, death=None)
#         self.assertIsNone(self.cm.set_alive(1, 11))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, death=None)}, self.cm._cluster)
#
#         self.assertIsNotNone(self.cm.set_alive(2, 21))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, death=None),
#                               2: _ClusterRegisterSession(2, session=21, birth=now, keepalive=now, death=None)}, self.cm._cluster)
#
#         self.assertIsNotNone(self.cm.set_alive(1, 12))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=12, birth=now, keepalive=now, death=None),
#                               2: _ClusterRegisterSession(2, session=21, birth=now, keepalive=now, death=None)},
#                              self.cm._cluster)
#
#     @mock.patch('dimensigon.utils.cluster_manager.get_now')
#     def test_set_death(self, mock_now):
#         mock_now.return_value = now
#
#         self.assertIsNotNone(self.cm.set_death(1, 11))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=now, death=now)}, self.cm._cluster)
#
#         self.cm._cluster[1] = _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=before_now, death=before_now)
#         self.assertIsNone(self.cm.set_death(1, 11))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=before_now, death=before_now)},
#                              self.cm._cluster)
#
#         self.cm._cluster[1] = _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=before_now, death=None)
#         self.assertIsNotNone(self.cm.set_death(1, 11))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=before_now, death=now)},
#                              self.cm._cluster)
#
#         self.cm._cluster[1] = _ClusterRegisterSession(1, session=11, birth=after_now, keepalive=after_now, death=None)
#         self.assertIsNotNone(self.cm.set_death(1, 11))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=after_now, keepalive=after_now, death=after_now)},
#                              self.cm._cluster)
#
#         self.cm._cluster[1] = _ClusterRegisterSession(1, session=11, birth=after_now, keepalive=after_now, death=None)
#         self.assertIsNotNone(self.cm.set_death(1, 11, before_now))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=after_now, keepalive=after_now, death=after_now)},
#                              self.cm._cluster)
#
#         self.cm._cluster[1] = _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=before_now, death=None)
#         self.assertIsNotNone(self.cm.set_death(1, 11, after_now))
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=before_now, death=after_now)},
#                              self.cm._cluster)
#
#         self.assertIsNone(self.cm.set_death(1, 12, after_now))
#
#     @mock.patch('dimensigon.utils.cluster_manager.get_now')
#     def test_set_keepalive(self, mock_now):
#         mock_now.return_value = now
#
#         self.cm.set_keepalive(1, 11)
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=now, keepalive=now, death=None)},
#                              self.cm._cluster)
#
#         self.cm._cluster[1] = _ClusterRegisterSession(1, session=11, birth=before_now, death=before_now)
#         self.cm.set_keepalive(1, 11)
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=now, death=None)},
#                              self.cm._cluster)
#
#         self.cm._cluster[1] = _ClusterRegisterSession(1, session=11, birth=before_now, death=None)
#         self.cm.set_keepalive(1, 11)
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=now, death=None)}, self.cm._cluster)
#
#         self.cm.set_keepalive(2, 21)
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=now, death=None),
#                               2: _ClusterRegisterSession(2, session=21, birth=now, keepalive=now, death=None)},
#                              self.cm._cluster)
#
#         self.cm.set_keepalive(2, 22)
#         self.assertDictEqual({1: _ClusterRegisterSession(1, session=11, birth=before_now, keepalive=now, death=None),
#                               2: _ClusterRegisterSession(2, session=22, birth=now, keepalive=now, death=None)},
#                              self.cm._cluster)
#
#     def test__update_cluster(self):
#         self.cm._cluster = {1: _ClusterRegisterSession(1, birth=before_now, keepalive=before_now, death=None),
#                             2: _ClusterRegisterSession(2, birth=now, keepalive=now, death=None),
#                             3: _ClusterRegisterSession(3, birth=before_now, keepalive=before_now, death=None),
#                             4: _ClusterRegisterSession(4, birth=before_now, keepalive=before_now, death=now),
#                             5: _ClusterRegisterSession(5, birth=before_now, keepalive=before_now, death=now),
#                             6: _ClusterRegisterSession(6, birth=before_now, keepalive=before_now, death=after_now),
#                             7: _ClusterRegisterSession(7, birth=before_now, keepalive=before_now, death=now),
#                             8: _ClusterRegisterSession(8, birth=before_now, keepalive=before_now, death=now),
#                             10: _ClusterRegisterSession(10, birth=before_now, keepalive=before_now, death=None)}
#
#         self.cm._update_cluster([_ClusterRegisterSession(1, birth=now, keepalive=now, death=None),
#                                  _ClusterRegisterSession(2, birth=before_now, keepalive=before_now, death=None),
#                                  _ClusterRegisterSession(3, birth=before_now, keepalive=before_now, death=now),
#                                  _ClusterRegisterSession(4, birth=before_now, keepalive=before_now, death=None),
#                                  _ClusterRegisterSession(5, birth=before_now, keepalive=before_now, death=after_now),
#                                  _ClusterRegisterSession(6, birth=before_now, keepalive=before_now, death=now),
#                                  _ClusterRegisterSession(7, birth=after_now, keepalive=after_now, death=None),
#                                  _ClusterRegisterSession(8, birth=after_now, keepalive=after_now, death=after_now),
#                                  _ClusterRegisterSession(9, birth=before_now, keepalive=before_now, death=None)])
#
#         self.assertDictEqual({1: _ClusterRegisterSession(1, birth=before_now, keepalive=now, death=None),
#                           2: _ClusterRegisterSession(2, birth=before_now, keepalive=now, death=None),
#                           3: _ClusterRegisterSession(3, birth=before_now, keepalive=before_now, death=now),
#                           4: _ClusterRegisterSession(4, birth=before_now, keepalive=before_now, death=now),
#                           5: _ClusterRegisterSession(5, birth=before_now, keepalive=before_now, death=now),
#                           6: _ClusterRegisterSession(6, birth=before_now, keepalive=before_now, death=now),
#                           7: _ClusterRegisterSession(7, birth=after_now, keepalive=after_now, death=None),
#                           8: _ClusterRegisterSession(8, birth=after_now, keepalive=after_now, death=after_now),
#                           9: _ClusterRegisterSession(9, birth=before_now, keepalive=before_now, death=None),
#                           10: _ClusterRegisterSession(10, birth=before_now, keepalive=before_now, death=None)},
#                          self.cm._cluster)


"""
usage: dshell [--version] [--username USER] [--password PASSWORD|--token TOKEN] [--server SERVER] [--port PORT]
              [--config-file CONFIG_FILE] [--debug] [COMMAND [ARGS...]]

options:
   -u, --username USER              Login user
   --password PASSWORD              Password used to login
   -t, --token TOKEN                Refresh token used to authenticate
   -s, --server SERVER              Server to communicate with
   -p, --port PORT                  Port to communicate with [default: 5000]
   -c, --config-file CONFIG_FILE    configuration file [default: ~/.dshell]
   --debug                          Sets trace for debugging

The available dshell commands are:
   status    Status of a node
   ping      Ping a node
   server    Server Resource
   orch      Orchestration Resource, run Orchestrations
   action    Action Resource
   exec      Execution Resource
   software  Software Resource, send software
   logfed    LogFed Resource
   manager   Commands to manage internal conditions

When no dshell command specified dshell starts in interactive mode

See 'dshell <command> --help' for more information on a specific command.

"""
import importlib
import logging
import os
import sys
from argparse import ArgumentParser

from docopt import docopt

import dimensigon.dshell.environ as env
from dimensigon.dshell.argparse_raise import create_parser
from dimensigon.dshell.bootstrap import bootstrap_environ
from dimensigon.dshell.commands import nested_dict
from dimensigon.dshell.interactive import interactive, call_func_with_signature

basename = os.path.dirname(os.path.abspath(__file__))

commands = 'cmd exec ping logfed manager server software status transfer refresh'.split()
batch_commands = 'action orch command'.split()  # commands that interact differently from interactive mode


def main():
    ch = logging.StreamHandler()
    logger = logging.getLogger('dshell')
    logger.addHandler(ch)

    # parse args
    args = docopt(__doc__,
                  version='dshell version 1.0',
                  options_first=True)

    bootstrap_environ(args)

    # process args
    argv = [args['COMMAND']] + args['ARGS']

    if not ('-h' or '--help') in argv:
        if args['COMMAND'] is None:
            import dimensigon.dshell
            dimensigon.dshell.INTERACTIVE = True
            interactive()
        else:
            if env._refresh_token is None:
                exit('No token specified. Unable to run command')
            if not env.get('SERVER', None):
                exit('No server specified. Unable to run command')
            if args['COMMAND'] in commands:
                parser = create_parser({args['COMMAND']: nested_dict[args['COMMAND']]},
                                       parser=ArgumentParser(prog="dshell"))
                namespace = parser.parse_args(argv)
                if hasattr(namespace, 'func'):
                    call_func_with_signature(vars(namespace))
            elif args['COMMAND'] in batch_commands:
                module = importlib.import_module('.dshell_%s' % args['COMMAND'], 'dimensigon.dshell.batch')
                module.main(argv)
            else:
                exit("%r is not a dshell command. See 'dshell --help'." % args['COMMAND'])
    else:
        parser = create_parser({args['COMMAND']: nested_dict[args['COMMAND']]},
                               parser=ArgumentParser(prog="dshell"))
        parser.parse_args(argv)

if __name__ == '__main__':
    sys.exit(main())



from setuptools import setup, find_packages

import dimensigon as dm

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='dimensigon',
    version=dm.__version__,
    packages=find_packages(),
    url='https://github.com/dimensigon/dimensigon',
    license=dm.__license__,
    author=dm.__author__,
    author_email=dm.__email__,
    description="Distributed Management and orchestration through RESTful, Mesh Networking and with a flair of IoT.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "PyYAML",
        "RestrictedPython",
        "aiohttp",
        "apscheduler",
        "click",
        "coolname",
        "coverage",
        "cryptography",
        "dataclasses",
        "dill",
        "docopt",
        "Flask",
        "flask-JWT-Extended",
        "flask-RESTFul",
        "flask-SQLAlchemy",
        "gunicorn",
        "jinja2",
        "jinja2schema",
        "jsonschema",
        "netifaces",
        "passlib",
        "prompt_toolkit",
        "pygments",
        "python-dateutil",
        "requests",
        "rsa",
        "schema",
        "setuptools",
        "six",
        "tzlocal<3.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX",
    ],
    python_requires='>=3.6',
    # scripts=['elevator.py'],
    entry_points={'console_scripts': ['dshell = dimensigon.dshell.batch.dshell:main',
                                      'dimensigon = dimensigon.__main__:main']},
)



import time

import aiohttp
import requests

from dimensigon.utils import asyncio


# def is_open(host: str, port: int, timeout: float = 1.0):
#     with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
#         sock.settimeout(timeout)
#         if sock.connect_ex((host, port)) == 0:
#             return True
#         else:
#             return False
#
#
# def is_ssl_open(host: str, port: int, timeout: float = 1.0):
#     context = mod_ssl.SSLContext()
#     try:
#         with socket.create_connection((host, port), timeout=timeout) as sock:
#             with context.wrap_socket(sock, server_hostname=host) as ssock:
#                 return True
#     except (ConnectionRefusedError, TimeoutError, OSError):
#         return False
#     except socket.timeout:
#         return False


def is_open2(host: str, port: int, timeout: float = 5.0, ssl=True):
    url = f"{'https' if ssl else 'http'}://{host}:{port}/"
    try:
        resp = requests.get(url, verify=False, timeout=timeout)
    except (requests.ConnectionError, requests.Timeout):
        return False
    else:
        return resp.ok


def check_host(host: str, port: int, retry=3, delay=2, timeout=5.0):
    ipup = False
    for i in range(retry):
        if is_open2(host, port, timeout):
            ipup = True
            break
        # elif is_open(host, port, timeout):
        #     ipup = True
        #     break
        time.sleep(delay)
    return ipup


###########
#  ASYNC  #
###########


# async def async_is_open(host: str, port: int, timeout: float = 1.0, ssl=True):
#     writer = None
#     kwargs = {}
#     if ssl:
#         kwargs['ssl'] = mod_ssl.SSLContext()
#     conn = asyncio.open_connection(host, port, **kwargs)
#     try:
#         reader, writer = await asyncio.wait_for(conn, timeout)
#     except (asyncio.TimeoutError, ConnectionRefusedError, OSError):
#         return False
#     finally:
#         if writer:
#             writer.close()
#     return True


async def async_is_open2(host: str, port: int, timeout: float = 5.0, ssl=True):
    url = f"{'https' if ssl else 'http'}://{host}:{port}/"
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=aiohttp.ClientTimeout(total=timeout), verify_ssl=False) as response:
                if 200 <= response.status < 300:
                    return True
                else:
                    return False
    except (aiohttp.ClientResponseError,
            aiohttp.ClientOSError,
            aiohttp.ServerDisconnectedError,
            aiohttp.ServerTimeoutError,
            asyncio.TimeoutError) as e:
        return False


async def async_check_host(host: str, port: int, retry=3, delay=2, timeout=5.0):
    ipup = False
    for i in range(retry):
        if await async_is_open2(host, port, timeout):
            ipup = True
            break
        # elif is_open(host, port, timeout):
        #     ipup = True
        #     break
        await asyncio.sleep(delay)
    return ipup



import datetime as dt
import json
import logging
import os
import random
import time
import typing as t

from apscheduler.schedulers.background import BackgroundScheduler
from flask import Flask, g
from flask_jwt_extended import JWTManager
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import MetaData

from dimensigon.utils.event_handler import EventHandler
from dimensigon.web import errors, threading
from dimensigon.web.config import config_by_name
from .extensions.flask_executor.executor import Executor
from .helpers import BaseQueryJSON, run_in_background
from .. import defaults
from ..utils import asyncio
from ..utils.helpers import get_now, bind2gate

if t.TYPE_CHECKING:
    from ..core import Dimensigon
    from ..use_cases.clustering import ClusterManager


meta = MetaData(naming_convention={
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
})

db = SQLAlchemy(query_class=BaseQueryJSON, metadata=meta,
                engine_options={'connect_args': {'check_same_thread': False}, 'pool_recycle': 60})
# db = SQLAlchemy(query_class=BaseQueryJSON, metadata=meta)
# db = SQLAlchemy(query_class=BaseQueryJSON)
jwt = JWTManager()
executor = Executor()


class DimensigonFlask(Flask):
    dm: t.ClassVar['Dimensigon'] = None
    cluster_manager: t.ClassVar['ClusterManager']

    def shutdown(self):
        with self.app_context():
            from dimensigon.domain.entities import Server, Parameter
            import dimensigon.web.network as ntwrk
            from dimensigon.use_cases.helpers import get_root_auth

            self.cluster_manager.stop()

            # shutdown executor
            executor.shutdown()
            if self.extensions.get('scheduler', None):
                self.extensions.get('scheduler').shutdown()

            Parameter.set('last_graceful_shutdown', get_now())
            db.session.commit()
            servers = Server.get_neighbours()
            if servers:
                self.logger.debug(f"Sending shutdown to {', '.join([s.name for s in servers])}")
            else:
                self.logger.debug("No server to send shutdown information")
            if servers:
                responses = asyncio.run(
                    ntwrk.parallel_requests(servers, 'post',
                                            view_or_url='api_1_0.cluster_out',
                                            view_data=dict(server_id=str(Server.get_current().id)),
                                            json={'death': get_now().strftime(defaults.DATEMARK_FORMAT)},
                                            timeout=2, auth=get_root_auth()))
                if self.logger.level <= logging.DEBUG:
                    for r in responses:
                        if not r.ok:
                            self.logger.warning(f"Unable to send data to {r.server}: {r}")

    def bootstrap_start(self):
        with self.app_context():
            from dimensigon.domain.entities import Server, Parameter, Route
            import dimensigon.web.network as ntwrk
            from dimensigon.use_cases.helpers import get_root_auth
            from dimensigon.use_cases import routing
            from dimensigon.domain.entities import Locker
            from ..use_cases.clustering import ClusterManager

            self.cluster_manager = ClusterManager(self,
                                                  threshold=self.dm.config.refresh_interval * defaults.COMA_NODE_FACTOR * 0.85,
                                                  start=False)

            # reset scopes
            Locker.set_initial(unlock=True)

            last_shutdown = Parameter.get('last_graceful_shutdown', defaults.INITIAL_DATEMARK) - dt.timedelta(hours=1)

            if self.dm.config.force_scan or last_shutdown < get_now() - self.dm.config.refresh_interval:
                scan = True
            else:
                scan = False
            asyncio.run(
                routing.async_update_routes_send(discover_new_neighbours=scan, check_current_neighbours=scan,
                                                 send=False))

            route_table = []
            for route in Route.query.join(Server.route).order_by(
                    Server.name).filter(Server.deleted == False).all():
                route_table.append(route.to_json(human=True))
            routing.logger.debug("Routing table: " + json.dumps(route_table, indent=2))

            # check gates
            me = Server.get_current()
            if me is None:
                raise RuntimeError("No server set as 'current'")

            input_gates = bind2gate(self.dm.config.http_conf.get('bind'))
            current_gates = [(gate.dns or str(gate.ip), gate.port) for gate in me.gates]
            new_gates = set(input_gates).difference(set(current_gates))
            self.server_id_with_new_gates = None
            if new_gates:
                if Parameter.get('join_server'):
                    join_server = Server.query.get(Parameter.get('join_server'))
                else:
                    join_server = None
                servers = Server.get_neighbours()
                if join_server in servers:
                    servers.pop(servers.index(join_server))
                    servers.append(join_server)
                else:
                    self.logger.warning(f'Join server {join_server} is not a neighbour')
                start = time.time()
                resp = None
                server = True
                while len(servers) > 0 and server and (time.time() - start) < 900:
                    server_retries = 0
                    server = servers[-1]
                    self.logger.debug(f"Sending new gates {new_gates} to {server}...")
                    resp = ntwrk.patch(server, 'api_1_0.serverresource',
                                       view_data=dict(server_id=str(Server.get_current().id)),
                                       json={'gates': [{'dns_or_ip': ip, 'port': port} for ip, port in new_gates]},
                                       timeout=60,
                                       auth=get_root_auth())
                    if not resp.ok:
                        self.logger.debug(f"Unable to send new gates to {server}. Reason: {resp}")
                        self.logger.info(f"Unable to create new gates. Trying to send again in 5 seconds...")
                        time.sleep(5)
                        if resp.code == 409:
                            # try with the same server
                            server_retries += 1
                        elif resp.code == 500:

                            # try with another server
                            i = servers.index(server) - 1
                            if i >= 0:
                                server = servers[i]
                                server_retries = 0
                            else:
                                server = None
                        if server_retries == 3:
                            # changing server
                            i = servers.index(server) - 1
                            if i >= 0:
                                server = servers[i]
                                server_retries = 0
                            else:
                                server = None
                    else:
                        self.logger.debug("New gates created succesfully")
                        self.server_id_with_new_gates = server.id
                        break

                if not servers:
                    if Server.query.count() == 1:
                        self.logger.info(f"Creating new gates {new_gates} without performing a lock on catalog")
                        for gate in new_gates:
                            g = me.add_new_gate(gate[0], gate[1])
                            db.session.add(g)
                        db.session.commit()
                else:
                    if resp and not resp.ok:
                        self.logger.warning(f"Remote servers may not connect with {me}. ")

        self.start_background_tasks()

    def make_first_request(self):
        from dimensigon.domain.entities import Server
        import dimensigon.web.network as ntwrk

        with self.app_context():
            start = time.time()
            while True:
                resp = ntwrk.get(Server.get_current(), 'root.home', timeout=1)
                if not resp.ok and time.time() - start < 30:
                    time.sleep(2)
                else:
                    break

    def notify_cluster(self):
        from dimensigon.domain.entities import Server
        import dimensigon.web.network as ntwrk
        from dimensigon.use_cases.helpers import get_root_auth
        from dimensigon.use_cases import routing
        from dimensigon.domain.entities import Parameter

        with self.app_context():
            cluster_logger = logging.getLogger('dimensigon.cluster')
            not_notify = set()
            me = Server.get_current()
            msg, debug_msg = routing.format_routes_message()

            neighbours = Server.get_neighbours()

            if Parameter.get('join_server'):
                join_server = Server.query.get(Parameter.get('join_server'))
            else:
                join_server = None

            if neighbours:
                random.shuffle(neighbours)
                first = [s for s in neighbours if s.id == self.server_id_with_new_gates]
                if first:
                    neighbours.pop(neighbours.index(first[0]))
                    neighbours = first + neighbours
                elif join_server in neighbours:
                        neighbours.pop(neighbours.index(join_server))
                        neighbours = [join_server] + neighbours
                for s in neighbours:
                    if s.id not in not_notify:
                        self.logger.debug(f"Sending 'Cluster IN' message to {s}")
                        resp = ntwrk.post(s, 'api_1_0.cluster_in', view_data=dict(server_id=str(me.id)),
                                          json=msg, timeout=10, auth=get_root_auth())
                        if resp.ok:
                            self.cluster_manager.cluster.update_cluster(resp.msg['cluster'])
                            not_notify.update(resp.msg.get('neighbours', []))
                        else:
                            self.logger.debug(f"Unable to send 'Cluster IN' message to {s} . Response: {resp}")
                    else:
                        self.logger.debug(f"Skiping server {s} from sending 'Cluster IN' message")
                else:
                    self.cluster_manager.set_alive(me.id)
                alive = [(getattr(Server.query.get(s_id), 'name', None) or s_id) for s_id in
                         self.cluster_manager.cluster.get_alive()]
                cluster_logger.info(f"Alive servers: {', '.join(alive)}")
            else:
                self.logger.debug("No neighbour to send 'Cluster IN'")
                self.cluster_manager.set_alive(me.id)

    def start_background_tasks(self):
        from ..use_cases.log_sender import LogSender
        from dimensigon.web.background_tasks import process_catalog_route_table
        if not self.config['TESTING'] or os.getenv('WERKZEUG_RUN_MAIN') == 'true':
            if self.extensions.get('scheduler') is None and self.config['SCHEDULER']:
                # passing timezone="UTC" to solve problems with systems where get_localzone() returns an object with
                # local zone
                bs = BackgroundScheduler(timezone="UTC")
                self.extensions['scheduler'] = bs
                ls = LogSender()
                self.extensions['log_sender'] = ls
                bs.start()

                # if self.config.get('AUTOUPGRADE'):
                # bs.add_job(func=process_get_new_version_from_gogs, args=(self,), trigger="interval",
                #            id='upgrader',
                #            hours=6)
                if self.dm.config.refresh_interval:
                    bs.add_job(func=process_catalog_route_table, name="catalog & route upgrade",
                               args=(self,),
                               id='routing_cluster_catalog_refresh',
                               trigger="interval", minutes=self.dm.config.refresh_interval.seconds / 60)
                bs.add_job(func=run_in_background, name="log_sender", args=(ls.send_new_data, self),
                           id='log_sender',
                           trigger="interval",
                           minutes=2, next_run_time=get_now() + dt.timedelta(seconds=30))

    def run(self, host=None, port=None, debug=None, **options):
        super(DimensigonFlask, self).run(host=host, port=port, debug=debug, use_reloader=False, **options)


def _initialize_blueprint(app):
    from dimensigon.web.routes import root_bp
    from dimensigon.web.api_1_0 import api_bp as api_1_0_bp

    app.register_blueprint(root_bp)
    handle_exception = app.handle_exception
    handle_user_exception = app.handle_user_exception
    app.register_blueprint(api_1_0_bp)
    app.handle_exception = handle_exception
    app.handle_user_exception = handle_user_exception


def _initialize_errorhandlers(app):
    '''
    Initialize error handlers
    '''
    from dimensigon.web.errors import bp_errors
    app.register_blueprint(bp_errors)


def create_app(config_name):
    app = DimensigonFlask('dimensigon')
    if isinstance(config_name, t.Mapping):
        app.config.from_mapping(config_name)
    elif config_name in config_by_name:
        app.config.from_object(config_by_name[config_name])
        config_by_name[config_name].init_app(app)
    else:
        app.config.from_object(config_name)
        if hasattr(config_name, 'init_app'):
            config_name.init_app(app)

    # EXTENSIONS
    db.init_app(app)
    jwt.init_app(app)
    executor.init_app(app)
    app.events = EventHandler()

    app.before_request(load_global_data_into_context)
    app.before_first_request(start_cluster_manager)
    _initialize_blueprint(app)
    _initialize_errorhandlers(app)

    return app


def start_cluster_manager():
    from flask import current_app
    current_app.notify_cluster()
    current_app.cluster_manager.start()


# @jwt.user_loader_callback_loader
# def user_loader_callback(identity):
#     from ..domain.entities import User
#     return User.query.get(identity)


def load_global_data_into_context():
    from dimensigon.domain.entities import Server, Dimension
    from dimensigon.web.decorators import set_source
    global _dimension, _server
    set_source()
    g.server = Server.get_current()
    g.dimension = Dimension.get_current()


from unittest import TestCase, mock

from flask import Flask
from flask_jwt_extended import JWTManager, create_access_token, jwt_required

from dimensigon.domain.entities import Server, Scope, Locker, State
from dimensigon.web import db
from dimensigon.web.decorators import lock_catalog


class TestSecurizer(TestCase):
    def setUp(self):
        """Create and configure a new self.app instance for each test."""
        # create a temporary file to isolate the database for each test
        # create the self.app with common test config

        self.app = Flask(__name__)
        self.app.config['JWT_SECRET_KEY'] = 'super-secret'

        self.jwt = JWTManager(self.app)

        @self.app.route('/', methods=['GET'])
        @jwt_required
        @lock_catalog
        def hello():
            return {'msg': 'default response'}

        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.init_app(self.app)
        db.create_all()
        # self.d = generate_dimension('test')
        self.srv1 = Server(id='bbbbbbbb-1234-5678-1234-56781234bbb1', name='server1', me=True)
        Locker.set_initial()
        db.session.add(self.srv1)
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @mock.patch('dimensigon.web.decorators.get_servers_from_scope')
    @mock.patch('dimensigon.web.decorators.lock_scope')
    def test_lock_catalog(self, mock_lock_scope, mock_get_servers_from_scope):
        mock_lock_scope.return_value.__enter__.return_value = 1
        mock_get_servers_from_scope.return_value = [self.srv1]
        resp = self.client.get('/',
                               headers={"Authorization": f"Bearer {create_access_token(1)}"})

        self.assertEqual(200, resp.status_code)
        mock_lock_scope.assert_called_once_with(Scope.CATALOG, [self.srv1])
        self.assertDictEqual({'msg': 'default response'}, resp.get_json())

    @mock.patch('dimensigon.web.decorators.get_servers_from_scope')
    @mock.patch('dimensigon.web.decorators.lock_scope')
    def test_lock_catalog_with_applicant(self, mock_lock_scope, mock_get_servers_from_scope):
        Locker.set_initial()
        l = Locker.query.get(Scope.CATALOG)
        l.state = State.LOCKED
        l.applicant = 2

        db.session.commit()

        mock_lock_scope.return_value.__enter__.return_value = 1
        mock_get_servers_from_scope.return_value = [self.srv1]
        resp = self.client.get('/',
                               headers={"Authorization": f"Bearer {create_access_token(1, user_claims={'applicant': 2})}"})

        self.assertEqual(200, resp.status_code)
        mock_lock_scope.assert_not_called()
        self.assertDictEqual({'msg': 'default response'}, resp.get_json())



_server2id_map = {}

_environ = {}

_access_token = None
_refresh_token = None
_username = None

def set_dict_in_environ(data: dict, **kwargs):
    for k, v in data.items():
        set(k, v)
    for k, v in kwargs.items():
        set(k, v)

def set(key, value):
    if key == 'ACCESS_TOKEN':
        raise ValueError('ACCESS_TOKEN is not settable')
    _environ.update({key: value})


def get(key, default=None):
    if key == 'ACCESS_TOKEN':
        return _access_token
    return _environ.get(key, default)


def set_server2id_map(data):
    for server in data:
        _server2id_map[server.get('name')] = server.get('id')


def server2id(server):
    return _server2id_map.get(server, None)




import json
import logging
import queue
import random
import threading
import typing as t
import uuid
from types import TracebackType
from typing import Optional, Type

from dimensigon.domain.entities import Server, Route, Gate
from dimensigon.domain.entities.route import RouteContainer
from dimensigon.network.low_level import check_host, async_check_host
from dimensigon.use_cases.helpers import get_root_auth
from dimensigon.utils import asyncio
from dimensigon.utils.helpers import convert, is_iterable_not_string, is_valid_uuid, format_exception
from dimensigon.utils.typos import Id
from dimensigon.web import network as ntwrk, db

logger = logging.getLogger('dimensigon.routing')


class _RLock(threading._RLock):

    def __enter__(self) -> bool:
        logger.log(1, "lock acquired")
        return super().__enter__()

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> Optional[bool]:
        db.session.commit()
        logger.log(1, "lock released")
        return super().__exit__(exc_type, exc_val, exc_tb)


_lock = _RLock()

MAX_COST = 99999


def check_gates(server, timeout=5, retries=1, delay=0.5) -> t.Union[RouteContainer, Route]:
    """checks if a server is neighbour

    Args:
        server: server to discover as a new neighbour
        timeout: timeout socket
        retries: number of times to try to connect to the socket

    Returns:
        returns Route if no change. This means that gate in inventory is still up.
                RouteContainer if there is an available gate to connect
                None if no gate to connect
    """

    # check current gate
    if server.route and server.route.gate:
        gate = server.route.gate
        if check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries, delay=delay, timeout=timeout):
            return server.route

    for gate in server.gates:
        if server.route and gate != server.route.gate:
            if (gate.ip and not gate.ip.is_loopback) or (gate.dns and gate.dns != 'localhost'):
                if check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries, delay=1, timeout=timeout):
                    return RouteContainer(None, gate, 0)


async def async_check_gates(server, timeout=5, retries=1, delay=1) -> t.Union[RouteContainer, Route]:
    """checks if a server is neighbour

    Args:
        server: server to discover as a new neighbour
        timeout: timeout socket
        retries: number of times to try to connect to the socket
        delay: time in seconds between retries

    Returns:
        returns Route if no change. This means that gate in inventory is still up.
                RouteContainer if there is an available gate to connect
                None if no gate to connect
    """

    # check current gate
    if server.route and server.route.gate:
        gate = server.route.gate
        if await async_check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries + 1, delay=delay,
                                  timeout=timeout):
            return server.route

    for gate in server.gates:
        if server.route and gate != server.route.gate:
            if (gate.ip and not gate.ip.is_loopback) or (gate.dns and gate.dns != 'localhost'):
                if await async_check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries,
                                          delay=1,
                                          timeout=timeout):
                    return RouteContainer(None, gate, 0)


def set_route(server: Server, route: RouteContainer):
    with _lock:
        server.set_route(route)
        db.session.add(server.route)
        db.session.commit()


def format_routes_message(routes: t.Dict[t.Union[Id, Server], RouteContainer] = None) -> t.Tuple[t.List, t.List]:
    if routes:
        iterator = routes.items()
    else:
        iterator = [(r.destination, r) for r in Route.query.all()]
    msg = []
    debug_msg = []
    for d, r in iterator:
        msg.append(
            dict(destination_id=getattr(d, 'id', d),
                 proxy_server_id=str(getattr(r.proxy_server, 'id')) if getattr(r.proxy_server, 'id',
                                                                               None) else None,
                 gate_id=str(getattr(r.gate, 'id')) if getattr(r.gate, 'id', None) else None,
                 cost=r.cost))

        debug_msg.append(dict(destination=getattr(d, 'name', d),
                              proxy_server=getattr(r.proxy_server, 'name') if getattr(r.proxy_server, 'name',
                                                                                      None) else None,
                              gate=str(r.gate) if r.gate else None,
                              cost=r.cost))

    return msg, debug_msg


async def async_send_routes(routes=None, auth=None, servers=None, exclude=None):
    if not servers:
        servers = Server.get_neighbours(exclude=exclude)
    msg, debug_msg = format_routes_message(routes)

    c_exclude = []
    if logger.level <= logging.DEBUG:
        if exclude:
            if is_iterable_not_string(exclude):
                c_exclude = [Server.query.get(e) if not isinstance(e, Server) else e for e in exclude]
            else:
                c_exclude = [Server.query.get(exclude) if not isinstance(exclude, Server) else exclude]
            log_msg = f" (Excluded nodes: {', '.join([getattr(e, 'name', e) for e in c_exclude])}):"
        else:
            log_msg = ''

        if servers:
            log_msg = f"Sending route information to the following nodes: {', '.join([s.name for s in servers])}" \
                      f"{log_msg}\n{json.dumps(debug_msg, indent=2)}"
        else:
            log_msg = f"No servers to send new routing information:{log_msg}\n{json.dumps(debug_msg, indent=2)}"

        logger.debug(log_msg)

    exclude_ids = list(set([s.id for s in servers]).union([getattr(e, 'id', e) for e in c_exclude]))

    aw = [ntwrk.async_patch(s, view_or_url='api_1_0.routes',
                            json={'server_id': Server.get_current().id, 'route_list': msg, 'exclude': exclude_ids},
                            auth=auth or get_root_auth()) for s
          in servers]

    rs = await asyncio.gather(*aw, return_exceptions=True)

    for r, s in zip(rs, servers):
        if isinstance(r, Exception):
            logger.warning(
                f"Error while trying to send route data to node {s}: "
                f"{format_exception(r)}")
        elif not r.ok:
            if r.exception:
                logger.warning(
                    f"Error while trying to send route data to node {s}: "
                    f"{format_exception(r.exception)}")
            else:
                logger.warning(f"Error while trying to send route data to node {s}: {r}")


async def _async_set_current_neighbours(neighbours: t.List[Server] = None,
                                        changed_routes: t.Dict[Server, RouteContainer] = None) -> t.List[Server]:
    """Function checks and sets neighbours

    Args:
        neighbours: list of neighbours
        changed_routes: reference to a dict which will be populated with new routes

    Returns:
        list of servers which are not neighbours anymore
    """
    not_neighbours_anymore = []

    if neighbours is None:
        neighbours = Server.get_neighbours()

    if neighbours:
        resp = await asyncio.gather(*[async_check_gates(server) for server in neighbours])
        for route, server in zip(resp, neighbours):
            if isinstance(route, RouteContainer):
                server.set_route(route)
                if changed_routes is not None:
                    changed_routes[server] = route
            elif route is None:
                not_neighbours_anymore.append(server)
                rc = RouteContainer(None, None, None)
                server.set_route(rc)
                if changed_routes is not None:
                    changed_routes[server] = rc
    return not_neighbours_anymore


async def _async_discover_new_neighbours(servers: t.List[Server] = None,
                                         changed_routes: t.Dict[Server, RouteContainer] = None) -> t.List[Server]:
    new_neighbours = []

    if servers is None:
        servers = Server.get_not_neighbours()

    if servers:
        resp = await asyncio.gather(*[async_check_gates(server) for server in servers], return_exceptions=True)
        for new_route, server in zip(resp, servers):
            if new_route:
                server.set_route(new_route)
                new_neighbours.append(server)
                changed_routes[server] = new_route

    return new_neighbours


def _route_table_merge(data: t.Dict[Server, ntwrk.Response]):
    changed_routes: t.Dict[Server, RouteContainer] = {}
    temp_table_routes: t.Dict[uuid.UUID, t.List[RouteContainer]] = {}
    me = Server.get_current()
    for s, resp in data.items():
        if resp.code == 200:
            server_id = resp.msg.get('server_id', None) or resp.msg.get('server').get('id')
            likely_proxy_server_entity = db.session.query(Server).get(server_id)
            for route_json in resp.msg['route_list']:
                route_json = convert(route_json)
                if route_json.destination_id != me.id \
                        and route_json.proxy_server_id != me.id \
                        and route_json.gate_id not in [g.id for g in me.gates]:
                    if route_json.destination_id not in temp_table_routes:
                        temp_table_routes.update({route_json.destination_id: []})
                    if route_json.cost is not None:
                        route_json.cost += 1
                        route_json.proxy_server_id = likely_proxy_server_entity.id
                        route_json.gate_id = None
                        temp_table_routes[route_json.destination_id].append(
                            RouteContainer(likely_proxy_server_entity.id, None, route_json.cost))
                    elif route_json.cost is None:
                        # remove a routing if gateway cannot reach the destination
                        temp_table_routes[route_json.destination_id].append(
                            RouteContainer(route_json.proxy_server_id, None, None))
        else:
            logger.error(f"Error while connecting with {s}. Error: {resp[1]}, {resp[0]}")

    # Select new routes based on neighbour routes
    neighbour_ids = [s.id for s in Server.get_neighbours()]
    for destination_id in filter(lambda s: s not in neighbour_ids, temp_table_routes.keys()):
        route = db.session.query(Route).filter_by(destination_id=destination_id).one_or_none()
        if not route:
            server = Server.query.get(destination_id)
            if not server:
                continue
            else:
                route = Route(destination=server)
        temp_table_routes[destination_id].sort(key=lambda x: x.cost or MAX_COST)
        if len(temp_table_routes[destination_id]) > 0:
            min_route = temp_table_routes[destination_id][0]
            proxy_server: Server = db.session.query(Server).get(min_route.proxy_server)
            cost = min_route.cost
            if route.proxy_server != proxy_server or route.cost != cost:
                rc = RouteContainer(proxy_server, None, cost)
                route.set_route(rc)
                changed_routes[route.destination] = rc
                db.session.add(route)

    data = {}
    for server, temp_route in changed_routes.items():
        data.update({str(server): {'proxy_server': str(temp_route.proxy_server), 'gate': str(temp_route.gate),
                                   'cost': str(temp_route.cost)}})
    return changed_routes


async def async_update_route_table_cost(discover_new_neighbours=False, check_current_neighbours=False, retries=2,
                                        timeout_conn=5, max_num_discovery=None) -> t.Dict[Server, RouteContainer]:
    """Gets route tables of all neighbours and updates its own table based on jump weights.
    Needs a Flask App Context to run.

    Parameters
    ----------
    discover_new_neighbours:
        tries to discover new neighbours
    check_current_neighbours:
        checks if current neighbours are still neighbours
    retries:
        number of times it will try to reach destination
    timeout_conn:
        time in seconds to stop waiting for connection

    Returns
    -------
    None
    """

    with _lock:
        logger.debug('Updating routing table')
        neighbours = Server.get_neighbours()
        not_neighbours = Server.get_not_neighbours()

        changed_routes: t.Dict[Server, RouteContainer] = {}

        not_neighbours_anymore = []
        new_neighbours = []

        aws = []
        if check_current_neighbours:
            if neighbours:
                logger.debug(f"Checking current neighbours: " + ', '.join([str(s) for s in neighbours]))
                aws.append(_async_set_current_neighbours(neighbours, changed_routes))
            else:
                logger.debug(f"No neighbour to check")

        if discover_new_neighbours:
            if not_neighbours[:max_num_discovery]:
                rs = list(not_neighbours)
                random.shuffle(rs)
                target = rs[:max_num_discovery]
                target.sort(key=lambda s: s.name)
                logger.debug(
                    f"Checking new neighbours{f' (limited to {max_num_discovery})' if max_num_discovery else ''}: "
                    + ', '.join([str(s) for s in target]))
                aws.append(_async_discover_new_neighbours(target, changed_routes))
            else:
                logger.debug("No new neighbours to check")

        res = await asyncio.gather(*aws, return_exceptions=False)

        if check_current_neighbours and neighbours:
            not_neighbours_anymore = res.pop(0)
            if not_neighbours_anymore:
                logger.info(
                    f"Lost direct connection to the following nodes: " + ', '.join(
                        [str(s) for s in not_neighbours_anymore]))
        if discover_new_neighbours and not_neighbours[:max_num_discovery]:
            new_neighbours = res.pop(0)
            if new_neighbours:
                logger.info(f'New neighbours found: ' + ', '.join([str(s) for s in new_neighbours]))
            else:
                logger.debug("No new neighbours found")

        # remove routes whose proxy_server is a node that is not a neighbour
        query = Route.query.filter(
            Route.proxy_server_id.in_([s.id for s in list(set(not_neighbours).union(set(not_neighbours_anymore)))]))
        rc = RouteContainer(None, None, None)
        for route in query.all():
            route.set_route(rc)
            changed_routes[route.destination] = rc

        # update neighbour list
        neighbours = list(set(neighbours).union(set(new_neighbours)) - set(not_neighbours_anymore))

        if neighbours:
            logger.debug(f"Getting routing tables from {', '.join([str(s) for s in neighbours])}")
            responses = await asyncio.gather(
                *[ntwrk.async_get(server, 'api_1_0.routes', auth=get_root_auth()) for server in neighbours])

            cr = _route_table_merge(dict(zip(neighbours, responses)))
            changed_routes.update(cr)

    return changed_routes


def update_route_table_from_data(new_routes: t.Dict, auth=None) -> t.Dict[Server, RouteContainer]:
    with _lock:
        likely_proxy_server = Server.query.get(new_routes.get('server_id'))
        changed_routes = {}
        me = Server.get_current()
        if not likely_proxy_server:
            logger.warning(f"Server id still '{new_routes.get('server_id')}' not created.")
            return changed_routes
        debug_new_routes = []
        r = new_routes.get('route_list', [])
        r.sort(key=lambda x: x.get('cost') or MAX_COST, reverse=True)
        for new_route in r:
            target_server = Server.query.get(new_route.get('destination_id'))
            proxy_server = Server.query.get(new_route.get('proxy_server_id'))
            gate = Gate.query.get(new_route.get('gate_id'))
            debug_new_route = dict(destination=getattr(target_server, 'name', new_route.get('destination_id')),
                                   proxy_server=getattr(proxy_server, 'name', new_route.get('proxy_server_id')),
                                   gate=str(gate) if gate else new_route.get('gate_id'),
                                   cost=new_route.get('cost'))
            debug_new_routes.append(debug_new_route)
            if target_server is None:
                logger.warning(f"Destination server unknown {new_route.get('destination_id')}")
                continue
            if target_server == me:
                # check if server has detected me as a neighbour
                if new_route.get('cost') == 0:
                    # check if I do not have it as a neighbour yet
                    if likely_proxy_server.route and likely_proxy_server.route.cost != 0:
                        # check if I have a gate to contact with it
                        route = check_gates(likely_proxy_server)
                        if isinstance(route, RouteContainer):
                            changed_routes[likely_proxy_server] = route
                            likely_proxy_server.set_route(route)
                        elif route is None:
                            if likely_proxy_server.route and likely_proxy_server.route.cost is not None:
                                changed_routes[likely_proxy_server] = RouteContainer(None, None, None)
                                likely_proxy_server.set_route(changed_routes[likely_proxy_server])
            else:
                # server may be created without route (backward compatibility)
                if target_server.route is None:
                    target_server.set_route(Route(destination=target_server))
                # process routes whose proxy_server is not me
                if str(me.id) != new_route.get('proxy_server_id'):
                    # check If I reach destination
                    if target_server.route.cost is not None:
                        if new_route.get('cost') is None:
                            #  likely proxy does not reach but I reach it. It might be shutdown unexpectedly?
                            if target_server.route.cost == 0:
                                # check if I still have it as a neighbour
                                route = check_gates(target_server)
                            else:
                                if target_server.route.proxy_server == likely_proxy_server:
                                    route = RouteContainer(None, None, None)
                                else:
                                    # check if I still have access through my proxy
                                    cost, time = ntwrk.ping(target_server, retries=1, timeout=20)
                                    if cost == target_server.route.cost:
                                        # still a valid route
                                        route = target_server.route
                                    elif cost is None:
                                        route = RouteContainer(None, None, None)
                                    else:
                                        route = RouteContainer(target_server.route.proxy_server, None, cost)

                            if isinstance(route, RouteContainer):
                                # gate changed
                                changed_routes[target_server] = route
                                target_server.set_route(route)
                            elif route is None:
                                # no route to host. I've lost contact too
                                changed_routes[target_server] = RouteContainer(None, None, None)
                                target_server.set_route(changed_routes[target_server])
                            else:
                                # still a valid route. Send route to likely_proxy_server to tell it I have access
                                resp = ntwrk.patch(likely_proxy_server, 'api_1_0.routes',
                                                   json=dict(server_id=str(me.id),
                                                             route_list=[target_server.route.to_json()]),
                                                   auth=auth or get_root_auth(), timeout=5)
                                if not resp.ok:
                                    logger.info(f'Unable to send route to {likely_proxy_server}: {resp}')
                        elif target_server.route.proxy_server is not None and \
                                target_server.route.proxy_server == likely_proxy_server:
                            # my proxy is telling me the route has changed
                            rc = RouteContainer(likely_proxy_server, None, new_route.get('cost') + 1)
                            target_server.set_route(rc)
                            changed_routes.update({target_server: rc})
                        elif new_route.get('cost') + 1 < target_server.route.cost:
                            # if new route has less cost than actual route, take it as my new route
                            rc = RouteContainer(likely_proxy_server, None, new_route.get('cost') + 1)
                            target_server.set_route(rc)
                            changed_routes.update({target_server: rc})
                    else:
                        # me does not reaches target_server
                        # if new route reaches the destination take it as a new one
                        if (new_route.get('cost') == 0 and new_route.get('gate_id') is not None) or (
                                new_route.get('cost') is not None and proxy_server is not None):
                            rc = RouteContainer(likely_proxy_server, None, new_route.get('cost') + 1)
                        else:
                            # neither my route and the new route has access to the destination
                            rc = RouteContainer(None, None, None)
                        if target_server.route and (
                                rc.proxy_server != target_server.route.proxy_server or
                                rc.gate != target_server.route.gate or
                                rc.cost != target_server.route.cost):
                            target_server.set_route(rc)
                            changed_routes.update({target_server: rc})
                else:
                    # target_server reached through me as a proxy from likely_proxy
                    pass

    query = Route.query.filter(
        Route.proxy_server_id.in_([s.id for s, r in changed_routes.items() if r.cost is None]))
    rc = RouteContainer(None, None, None)
    for route in query.all():
        route.set_route(rc)
        changed_routes[route.destination] = rc

    logger.debug(
        f"New routes processed from {likely_proxy_server.name}: {json.dumps(debug_new_routes, indent=2)}")
    # if changed_routes:
    #     Parameter.set('routing_last_refresh', get_now())
    return changed_routes


async def async_update_routes_send(discover_new_neighbours=False, check_current_neighbours=False,
                                   send=True, send_full_routes=False, max_num_discovery=None,
                                   retries=2, timeout_conn=5):
    # first, update route table
    changed_routes = await async_update_route_table_cost(discover_new_neighbours=discover_new_neighbours,
                                                         check_current_neighbours=check_current_neighbours,
                                                         retries=retries,
                                                         timeout_conn=timeout_conn,
                                                         max_num_discovery=max_num_discovery)
    # send route information
    if send and len(changed_routes) > 0 or send_full_routes:
        await async_send_routes(False if send_full_routes else changed_routes, get_root_auth())


async def async_remove_neighbour_send(server: t.Union[Id, Server], auth, servers=None):
    if not isinstance(server, Server):
        _server = Server.query.get(server)
    else:
        _server = server

    changed_routes = {}

    if _server:
        with _lock:
            _server.set_route(RouteContainer(None, None, None))
            lost_routes = Route.query.filter_by(proxy_server=_server).count()
            if lost_routes:
                changed_routes = await async_update_route_table_cost(discover_new_neighbours=False,
                                                                     check_current_neighbours=False)
    changed_routes.update({_server or server: RouteContainer(None, None, None)})
    await async_send_routes(changed_routes, auth, servers=servers)


async def async_check_set_neighbour_send(server: t.Union[Id, Server], auth):
    with _lock:
        if not isinstance(server, Server):
            if is_valid_uuid(server):
                server = Server.query.get(server)
            else:
                return
        else:
            if server not in db.session:
                server = db.session.merge(server, load=True)

        if server and server.route and server.route.cost != 0:
            new_route = await async_check_gates(server)
            if isinstance(new_route, RouteContainer):
                logger.debug(f"New neighbour {server} found on keepalive with gate {new_route.gate}")
                server.set_route(new_route)
                await async_send_routes({server: new_route}, exclude=server, auth=auth)


class RouteManager:

    def __init__(self, app, maxsize=None, retain_time=2, start=True):
        self.app = app
        self.queue = queue.Queue(maxsize=maxsize or 10000)
        self._stop = False
        self.buffer: t.Dict[Id, t.Dict] = {}
        self.retain_time = retain_time
        self.loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self.run, name='route_manager')
        self._change_buffer_lock = threading.RLock()
        self._timer = None
        if start:
            self._thread.start()

    def start(self):
        self._thread.start()

    def stop(self):
        self._stop = True
        self.queue.put('STOP', block=False)
        if self._timer:
            self._timer.cancel()

    def _add2buffer(self, data: t.Union[t.List[t.Dict], t.Dict]):
        with self._change_buffer_lock:
            if isinstance(data, dict):
                self.buffer.update({data['id']: data})
            else:
                self.buffer.update(**{cr['id']: cr for cr in data})
            if self._timer is None:
                self._timer = threading.Timer(interval=self.retain_time, function=self.send_data)
                self._timer.start()

    def put(self, data: t.Union[t.Dict, t.List[t.Dict]], block=True, timeout=None):
        self.queue.put(data, block=block, timeout=timeout)

    def send_data(self):
        with self.app.app_context():
            # time to send data
            neighbours = Server.get_neighbours()
            if neighbours:
                logger.debug(
                    f"Sending route information to the following nodes {', '.join([s.name for s in neighbours])}")

                auth = get_root_auth()
                with self._change_buffer_lock:
                    temp_buffer = dict(self.buffer)
                    self.buffer.clear()
                if logger.level <= logging.DEBUG:
                    logger.debug(f"{json.dumps(temp_buffer.values(), indent=2)}")
                try:
                    responses = self.loop.run_until_complete(
                        ntwrk.parallel_requests(neighbours, 'POST', view_or_url='api_1_0.cluster',
                                                json=list(temp_buffer.values()), auth=auth))
                except Exception as e:
                    logger.error(f"Unable to send route information to neighbours: "
                                 f"{format_exception(e)}")
                    # restore data with new data arrived
                    with self._change_buffer_lock:
                        temp_buffer.update(**self.buffer)
                        self.buffer.clear()
                        self.buffer.update(temp_buffer)
                else:
                    for s, r in zip(neighbours, responses):
                        if not r.ok:
                            logger.warning(f"Unable to send data to {s}: {r}")

                # set timer to launch again on change
                with self._change_buffer_lock:
                    self._timer = None
            else:
                logger.debug(f"No neighbour servers to send route information")
                with self._change_buffer_lock:
                    self._timer = None

    def run(self):
        logger.debug('Starting route manager')
        while not self._stop:
            try:
                item = self.queue.get(block=True, timeout=5)
            except queue.Empty:
                pass
            else:
                if item == 'STOP':
                    break
                elif item:
                    pass
        logger.debug('Stopping route manager')